"""
Module to provide processing for the handling of leaf elements within container blocks.
"""

from __future__ import annotations

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.leaf_blocks.atx_leaf_block_processor import AtxLeafBlockProcessor
from pymarkdown.leaf_blocks.fenced_leaf_block_processor import FencedLeafBlockProcessor
from pymarkdown.leaf_blocks.indented_leaf_block_processor import (
    IndentedLeafBlockProcessor,
)
from pymarkdown.leaf_blocks.leaf_block_processor import LeafBlockProcessor
from pymarkdown.leaf_blocks.leaf_block_processor_paragraph import (
    LeafBlockProcessorParagraph,
)
from pymarkdown.leaf_blocks.setext_leaf_block_processor import SetextLeafBlockProcessor
from pymarkdown.leaf_blocks.table_block_processor import TableBlockHelper
from pymarkdown.leaf_blocks.thematic_leaf_block_processor import (
    ThematicLeafBlockProcessor,
)
from pymarkdown.links.link_reference_definition_helper import (
    LinkReferenceDefinitionHelper,
)
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import ListStackToken

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-many-lines
# pylint: disable=too-few-public-methods


class ContainerBlockLeafProcessor:
    """
    Class to provide processing for the handling of leaf elements within container blocks.
    """

    @staticmethod
    def handle_leaf_tokens(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        """
        Handle the handoff between processing containers and the leaf tokens.
        """
        assert (
            parser_state.original_line_to_parse is not None
        ), "Original line should be defined by this point."
        calculated_indent = len(parser_state.original_line_to_parse) - len(
            grab_bag.line_to_parse
        )
        # POGGER.debug(">>indent>>$", calculated_indent)

        assert not (
            grab_bag.indent_used_by_list
            and ">" in grab_bag.indent_used_by_list
            and parser_state.token_stack[-1].is_paragraph
            and parser_state.token_stack[-2].is_block_quote
        ), "This sequence should have been dealt with before here."

        POGGER.debug("ttp>>:$:<", position_marker.text_to_parse)
        POGGER.debug("index_number>>:$:<", position_marker.index_number)
        POGGER.debug("index_indent>>:$:<", position_marker.index_indent)
        newer_position_marker = PositionMarker(
            position_marker.line_number,
            0,
            grab_bag.line_to_parse,
            index_indent=calculated_indent,
        )
        POGGER.debug("ttp>>:$:<", newer_position_marker.text_to_parse)
        POGGER.debug("index_number>>:$:<", newer_position_marker.index_number)
        POGGER.debug("index_indent>>:$:<", newer_position_marker.index_indent)
        parser_state.mark_for_leaf_processing(grab_bag.container_tokens)

        top_of_stack = parser_state.token_stack[-1]
        ContainerBlockLeafProcessor.__process_leaf_tokens(
            parser_state,
            newer_position_marker,
            grab_bag,
        )
        if grab_bag.requeue_line_info:
            parser_state.abc(grab_bag.requeue_line_info, top_of_stack)

        parser_state.clear_after_leaf_processing()

        grab_bag.extend_container_tokens_with_leaf_tokens()

    @staticmethod
    def __process_leaf_tokens(
        parser_state: ParserState,
        xposition_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        """
        Main entry point from the ContainerBlockProcessor for processing leaf tokens.
        """
        assert (
            grab_bag.is_leaf_tokens_empty()
        ), "No leaf tokens should be present at this point."
        POGGER.debug("parsing leaf>>")
        # if xposition_marker.index_number != 0 and grab_bag.removed_chars_at_start_of_line:
        #     ff = False
        #     assert ff
        position_marker = PositionMarker(
            xposition_marker.line_number,
            0,
            xposition_marker.text_to_parse,
            index_indent=xposition_marker.index_indent,
        )
        # POGGER.debug("position_marker.text>>:$:<<", position_marker.text_to_parse)
        # POGGER.debug("position_marker.index>>:$:<<", position_marker.index_number)
        # POGGER.debug("position_marker.indent>>:$:<<", position_marker.index_indent)

        (
            new_index_number,
            extracted_leaf_whitespace,
        ) = ParserHelper.extract_spaces_verified(position_marker.text_to_parse, 0)

        total_ws = new_index_number + position_marker.index_indent

        last_block_index = parser_state.find_last_block_quote_on_stack()
        last_list_index = parser_state.find_last_list_block_on_stack()

        ContainerBlockLeafProcessor.__handle_special_block_quote_reduction(
            parser_state,
            last_block_index,
            last_list_index,
            grab_bag,
        )
        if grab_bag.requeue_line_info:
            POGGER.debug("requeuing after __handle_special_block_quote_reduction")
            assert (
                grab_bag.is_leaf_tokens_empty()
            ), "No leaf tokens should be present at this point."
            return

        POGGER.debug("position_marker.text>>:$:<<", position_marker.text_to_parse)
        POGGER.debug("position_marker.index>>:$:<<", position_marker.index_number)
        POGGER.debug("position_marker.indent>>:$:<<", position_marker.index_indent)
        (
            adjust_token,
            position_marker,
            extracted_leaf_whitespace,
            grab_bag.text_removed_by_container,
        ) = ContainerBlockLeafProcessor.__adjust_for_inner_list_container(
            parser_state,
            last_block_index,
            last_list_index,
            position_marker,
            grab_bag.text_removed_by_container,
            extracted_leaf_whitespace,
        )
        orig_text_removed_by_container = grab_bag.text_removed_by_container

        POGGER.debug("position_marker.text>>:$:<<", position_marker.text_to_parse)
        POGGER.debug("position_marker.index>>:$:<<", position_marker.index_number)
        POGGER.debug("position_marker.indent>>:$:<<", position_marker.index_indent)
        (
            removed_leading_space,
            actual_removed_leading_space,
            position_marker,
            new_ex,
        ) = ContainerBlockLeafProcessor.__adjust_for_list_container(
            parser_state,
            position_marker,
            last_block_index,
            last_list_index,
            extracted_leaf_whitespace,
            grab_bag,
        )

        POGGER.debug("position_marker.text>>:$:<<", position_marker.text_to_parse)
        POGGER.debug("position_marker.index>>:$:<<", position_marker.index_number)
        POGGER.debug("position_marker.indent>>:$:<<", position_marker.index_indent)
        POGGER.debug("is_leaf_tokens_empty>>:$:<<", grab_bag.is_leaf_tokens_empty())
        POGGER.debug(
            "do_skip_containers_before_leaf_blocks>>:$:<<",
            grab_bag.do_skip_containers_before_leaf_blocks,
        )
        POGGER.debug("removed_leading_space>>:$:<<", removed_leading_space)
        if (
            grab_bag.is_leaf_tokens_empty()
            and not grab_bag.do_skip_containers_before_leaf_blocks
            and not removed_leading_space
        ):
            position_marker = (
                ContainerBlockLeafProcessor.__adjust_containers_before_leaf_blocks(
                    parser_state,
                    position_marker,
                    last_block_index,
                    total_ws,
                    actual_removed_leading_space,
                    grab_bag,
                    new_ex,
                )
            )
            # POGGER.debug("position_marker.text>>:$:<<", position_marker.text_to_parse)
            # POGGER.debug("position_marker.index>>:$:<<", position_marker.index_number)

        POGGER.debug("position_marker.text>>:$:<<", position_marker.text_to_parse)
        POGGER.debug("position_marker.index>>:$:<<", position_marker.index_number)
        POGGER.debug("position_marker.indent>>:$:<<", position_marker.index_indent)
        ContainerBlockLeafProcessor.__parse_line_for_leaf_blocks(
            parser_state,
            position_marker,
            grab_bag,
        )

        ContainerBlockLeafProcessor.__process_leaf_tokens_adjust(adjust_token, grab_bag)

        ContainerBlockLeafProcessor.__post_leaf_block_adjustment(
            parser_state,
            orig_text_removed_by_container,
            position_marker.line_number,
        )

    @staticmethod
    def __process_leaf_tokens_adjust(
        adjust_token: Optional[ListStartMarkdownToken], grab_bag: ContainerGrabBag
    ) -> None:
        if (
            adjust_token is not None
            and grab_bag.leaf_tokens
            and grab_bag.leaf_tokens[0].is_end_token
        ) and grab_bag.leaf_tokens[0].is_block_quote_end:
            POGGER.debug(
                "__process_leaf_tokens>>adjust_token>>$",
                adjust_token,
            )
            adjust_token.remove_last_leading_space()
            POGGER.debug(
                "__process_leaf_tokens>>adjust_token>>$",
                adjust_token,
            )

            token_index = 1
            end_container_count = 0
            end_list_count = 0
            while token_index < len(grab_bag.leaf_tokens):
                if (
                    grab_bag.leaf_tokens[token_index].is_block_quote_end
                    or grab_bag.leaf_tokens[token_index].is_list_end
                ):
                    end_container_count += 1
                if grab_bag.leaf_tokens[token_index].is_list_end:
                    end_list_count += 1
                token_index += 1
            if end_container_count != 0 and token_index != 0 and grab_bag.bogus:
                POGGER.debug(
                    "__process_leaf_tokens>>adjust_token>>$",
                    adjust_token,
                )
                adjust_token.remove_last_leading_space()
                POGGER.debug(
                    "__process_leaf_tokens>>adjust_token>>$",
                    adjust_token,
                )

                POGGER.debug(
                    "__process_leaf_tokens>>adjust_token>>$",
                    adjust_token,
                )
                adjust_token.add_leading_spaces(grab_bag.bogus)
                POGGER.debug(
                    "__process_leaf_tokens>>adjust_token>>$",
                    adjust_token,
                )

    @staticmethod
    def __parse_line_for_leaf_blocks(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> None:
        """
        Parse the contents of a line for a leaf block.

        Note: This is one of the more heavily traffic functions in the
        parser.  Debugging should be uncommented only if needed.
        """
        POGGER.debug("parsing leaf tokens")
        POGGER.debug("Leaf Line:$:", position_marker.text_to_parse)
        POGGER.debug(
            "Indent:$:, Index:$:",
            position_marker.index_indent,
            position_marker.index_number,
        )
        POGGER.debug("original_line:$:", grab_bag.original_line)
        detabified_original_line = TabHelper.detabify_string(grab_bag.original_line)
        detabified_original_start_index = detabified_original_line.find(
            position_marker.text_to_parse
        )
        assert (
            detabified_original_start_index != -1
        ), "Text must also be found in the original line."
        POGGER.debug(
            "detabified_original_start_index:$:", detabified_original_start_index
        )

        (
            outer_processed,
            leaf_block_position_marker,
            leaf_token_whitespace,
        ) = ContainerBlockLeafProcessor.__handle_block_leaf_tokens(
            parser_state, position_marker, detabified_original_start_index, grab_bag
        )
        if not outer_processed:
            assert (
                grab_bag.removed_chars_at_start_of_line is not None
            ), "When processing these leaf elements, removed_chars_at_start_of_line must be defined by now."
            new_tokens = (
                AtxLeafBlockProcessor.parse_atx_headings(
                    parser_state,
                    leaf_block_position_marker,
                    leaf_token_whitespace,
                    grab_bag.block_quote_data,
                    grab_bag.original_line,
                    grab_bag,
                )
                or IndentedLeafBlockProcessor.parse_indented_code_block(
                    parser_state,
                    leaf_block_position_marker,
                    leaf_token_whitespace,
                    grab_bag.removed_chars_at_start_of_line,
                    grab_bag.last_block_quote_index,
                    grab_bag.original_line,
                )
                or SetextLeafBlockProcessor.parse_setext_headings(
                    parser_state,
                    leaf_block_position_marker,
                    leaf_token_whitespace,
                    grab_bag.block_quote_data,
                    grab_bag.original_line,
                )
                or ThematicLeafBlockProcessor.parse_thematic_break(
                    parser_state,
                    leaf_block_position_marker,
                    leaf_token_whitespace,
                    grab_bag.block_quote_data,
                    grab_bag.original_line,
                    grab_bag,
                )
                or LeafBlockProcessorParagraph.parse_paragraph(
                    parser_state,
                    leaf_block_position_marker,
                    leaf_token_whitespace,
                    grab_bag.block_quote_data,
                    grab_bag.text_removed_by_container,
                    grab_bag.original_line,
                )
            )
            # POGGER.debug(">>leaf--adding>>$", new_tokens)
            grab_bag.extend_leaf_tokens(new_tokens)

    @staticmethod
    def __handle_block_leaf_tokens(
        parser_state: ParserState,
        incoming_position_marker: PositionMarker,
        detabified_original_start_index: int,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[bool, PositionMarker, str]:
        POGGER.debug(
            "line>>$>>index>>$>>",
            incoming_position_marker.text_to_parse,
            incoming_position_marker.index_number,
        )
        remaining_line_to_parse = incoming_position_marker.text_to_parse[
            incoming_position_marker.index_number :
        ]
        (new_index_number, leaf_token_whitespace) = (
            ParserHelper.extract_spaces_verified(
                incoming_position_marker.text_to_parse,
                incoming_position_marker.index_number,
            )
        )
        POGGER.debug(">>leaf_token_whitespace>>:$:<<", leaf_token_whitespace)

        position_marker = PositionMarker(
            incoming_position_marker.line_number,
            new_index_number,
            incoming_position_marker.text_to_parse,
            index_indent=incoming_position_marker.index_indent,
        )

        pre_tokens = LeafBlockProcessor.close_indented_block_if_indent_not_there(
            parser_state, leaf_token_whitespace
        )

        new_tokens: List[MarkdownToken] = []
        outer_processed = FencedLeafBlockProcessor.handle_fenced_code_block(
            parser_state,
            position_marker,
            leaf_token_whitespace,
            new_tokens,
            grab_bag.original_line,
            detabified_original_start_index,
            grab_bag.block_quote_data,
            grab_bag,
        )

        ignore_lrd_start = (
            grab_bag.do_ignore_link_definition_start
            or parser_state.token_stack[-1].is_html_block
        )
        ignore_table_start = grab_bag.do_ignore_table_start

        (
            outer_processed,
            grab_bag.requeue_line_info,
        ) = LinkReferenceDefinitionHelper.handle_link_reference_definition_leaf_block(
            parser_state,
            outer_processed,
            position_marker,
            leaf_token_whitespace,
            remaining_line_to_parse,
            ignore_lrd_start,
            pre_tokens,
            grab_bag.original_line,
        )

        (
            outer_processed,
            grab_bag.requeue_line_info,
        ) = TableBlockHelper.handle_table_leaf_block(
            parser_state,
            outer_processed,
            position_marker,
            leaf_token_whitespace,
            remaining_line_to_parse,
            ignore_table_start,
            pre_tokens,
            grab_bag.original_line,
            grab_bag.requeue_line_info,
        )

        outer_processed = LeafBlockProcessor.handle_html_block(
            parser_state,
            position_marker,
            outer_processed,
            leaf_token_whitespace,
            new_tokens,
            grab_bag,
        )
        grab_bag.extend_leaf_tokens(pre_tokens)
        grab_bag.extend_leaf_tokens(new_tokens)
        return (
            outer_processed,
            position_marker,
            leaf_token_whitespace,
        )

    @staticmethod
    def __handle_special_block_quote_reduction(
        parser_state: ParserState,
        last_block_index: int,
        last_list_index: int,
        grab_bag: ContainerGrabBag,
    ) -> None:
        assert (
            grab_bag.is_leaf_tokens_empty()
        ), "No leaf tokens should be present at this point."
        real_stack_count = parser_state.count_of_block_quotes_on_stack()
        POGGER.debug(
            "stack_count>>:$:($)<",
            grab_bag.block_quote_data.stack_count,
            real_stack_count,
        )
        POGGER.debug("??? special_block_quote_reduction")
        if (
            grab_bag.block_quote_data.current_count < real_stack_count
            and not grab_bag.was_paragraph_continuation
            and last_block_index > last_list_index
            and not last_list_index
        ):
            POGGER.debug("yes special_block_quote_reduction")
            stack_delta = (
                grab_bag.block_quote_data.stack_count
                - grab_bag.block_quote_data.current_count
            )
            if parser_state.token_stack[-1].was_link_definition_started:
                stack_delta += 1
            (
                close_tokens,
                grab_bag.requeue_line_info,
            ) = parser_state.close_open_blocks_fn(
                parser_state,
                include_block_quotes=True,
                was_forced=True,
                requeue_reset=True,
                caller_can_handle_requeue=True,
                until_this_index=len(parser_state.token_stack) - stack_delta,
            )
            grab_bag.extend_leaf_tokens(close_tokens)
            if not grab_bag.requeue_line_info:
                POGGER.debug("parser_state.token_stack>>$", parser_state.token_stack)
                POGGER.debug(
                    "parser_state.token_document>>$", parser_state.token_document
                )
                grab_bag.block_quote_data = BlockQuoteData(
                    grab_bag.block_quote_data.current_count,
                    grab_bag.block_quote_data.stack_count - stack_delta,
                )

    @staticmethod
    def __xx1(
        parser_state: ParserState,
        position_marker: PositionMarker,
        text_removed_by_container: str,
        extracted_leaf_whitespace: str,
    ) -> Tuple[PositionMarker, str, str, str]:
        stack_index = 1
        removed_text_copy = text_removed_by_container[:]
        removed_text_copy_bq_count = removed_text_copy.count(">")
        bq_count = 0
        while bq_count < removed_text_copy_bq_count:
            if parser_state.token_stack[stack_index].is_block_quote:
                bq_count += 1
            stack_index += 1
        last_bq_char_index = removed_text_copy.rindex(">")
        last_bq_char_index += 1
        assert removed_text_copy[last_bq_char_index] == " "
        last_bq_char_index += 1
        assert last_bq_char_index == len(removed_text_copy)

        ws_to_use = 0
        dd = parser_state.token_stack[stack_index].matching_markdown_token
        assert dd is not None
        assert dd.is_list_start
        dd_list = cast(ListStartMarkdownToken, dd)
        indent_delta = dd_list.indent_level - len(removed_text_copy)
        if len(extracted_leaf_whitespace) >= indent_delta:
            ws_to_use += indent_delta

        if ws_to_use:
            ex_space = (
                extracted_leaf_whitespace[:ws_to_use] + ParserLogger.blah_sequence
            )
            extracted_leaf_whitespace = extracted_leaf_whitespace[ws_to_use:]
            new_position_marker = PositionMarker(
                line_number=position_marker.line_number,
                index_number=position_marker.index_number,
                text_to_parse=position_marker.text_to_parse[ws_to_use:],
                index_indent=position_marker.index_indent + ws_to_use,
            )
            position_marker = new_position_marker
            text_removed_by_container += ex_space
        else:
            ex_space = ""
        return (
            position_marker,
            extracted_leaf_whitespace,
            text_removed_by_container,
            ex_space,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_inner_list_container(
        parser_state: ParserState,
        last_block_index: int,
        last_list_index: int,
        position_marker: PositionMarker,
        text_removed_by_container: Optional[str],
        extracted_leaf_whitespace: str,
    ) -> Tuple[Optional[ListStartMarkdownToken], PositionMarker, str, Optional[str]]:
        POGGER.debug("??? adjust_for_inner_list_container")
        if last_block_index > 0 and 0 < last_list_index < last_block_index:
            POGGER.debug("yes adjust_for_inner_list_container")
            list_token = cast(
                ListStartMarkdownToken,
                parser_state.token_stack[last_list_index].matching_markdown_token,
            )
            if list_token.line_number != position_marker.line_number:
                POGGER.debug("plt-a>>last_block_token>>$", list_token)

                ex_space = ""
                if text_removed_by_container is not None and text_removed_by_container:
                    lt_indent = list_token.indent_level
                    orig_ws_len = len(text_removed_by_container)
                    if orig_ws_len < lt_indent and extracted_leaf_whitespace:

                        (
                            position_marker,
                            extracted_leaf_whitespace,
                            text_removed_by_container,
                            ex_space,
                        ) = ContainerBlockLeafProcessor.__xx1(
                            parser_state,
                            position_marker,
                            text_removed_by_container,
                            extracted_leaf_whitespace,
                        )

                POGGER.debug(
                    "__adjust_for_inner_list_container>>list_token>>$", list_token
                )
                list_token.add_leading_spaces(ex_space)
                POGGER.debug(
                    "__adjust_for_inner_list_container>>list_token>>$", list_token
                )
                return (
                    list_token,
                    position_marker,
                    extracted_leaf_whitespace,
                    text_removed_by_container,
                )
        else:
            POGGER.debug("not adjust_for_inner_list_container")
        return (
            None,
            position_marker,
            extracted_leaf_whitespace,
            text_removed_by_container,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __adjust_for_list_container_after_block_quote_special_special(
        parser_state: ParserState,
        adj_original: str,
        bq_index: int,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[str, str]:
        assert (
            grab_bag.text_removed_by_container is not None
        ), "Block quote processing means we removed at least some text."
        removed_text_length = len(grab_bag.text_removed_by_container)
        if (
            adj_original[removed_text_length - 1] == "\t"
            and adj_original[removed_text_length - 2] == ">"
        ):
            orig_prefix = adj_original[removed_text_length - 1 :]
            orig_suffix = adj_original[: removed_text_length - 1]
            xx_block_quote_token = cast(
                BlockQuoteMarkdownToken,
                parser_state.token_stack[bq_index].matching_markdown_token,
            )
            _ = xx_block_quote_token.remove_last_bleading_space()
            # assert (
            #     last_leading_space[0] == "\n"
            # ), "Removed leading space must start with \\n."
            # last_leading_space = last_leading_space[1:]

            POGGER.debug(
                "__adjust_for_list_container_after_block_quote_special_special>>block_token>>$",
                xx_block_quote_token,
            )
            xx_block_quote_token.add_bleading_spaces(">")
            POGGER.debug(
                "__adjust_for_list_container_after_block_quote_special_special>>block_token>>$",
                xx_block_quote_token,
            )
        else:
            orig_prefix = adj_original[removed_text_length:]
            orig_suffix = adj_original[:removed_text_length]
        (
            non_space_index,
            ex_ws,
        ) = ParserHelper.extract_spaces_verified(orig_prefix, 0)
        original_up_to_non_space = orig_prefix[:non_space_index]
        redone_original = orig_suffix + original_up_to_non_space
        detabified_redone_original = TabHelper.detabify_string(redone_original)

        return detabified_redone_original, ex_ws

    @staticmethod
    def __adjust_for_list_container_after_block_quote_special(
        parser_state: ParserState,
        xposition_marker: PositionMarker,
        extracted_leaf_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> str:
        bq_index = parser_state.find_last_block_quote_on_stack()
        list_index = parser_state.find_last_list_block_on_stack()

        current_list_indent = -1
        assert (
            list_index > bq_index
        ), "If here, the list is embedded within a block quotes, and the indices must support that."
        xx_list_token = cast(
            ListStartMarkdownToken,
            parser_state.token_stack[list_index].matching_markdown_token,
        )
        current_list_indent = xx_list_token.indent_level

        assert (
            grab_bag.text_removed_by_container is not None
        ), "If here, some text must have been removed."
        reconstructed_line = (
            grab_bag.text_removed_by_container + xposition_marker.text_to_parse
        )
        (
            adj_original,
            _,
            _,
        ) = TabHelper.find_tabified_string(
            grab_bag.original_line,
            reconstructed_line,
            abc=False,
            was_indented=True,
            reconstruct_prefix=None,
        )
        (
            detabified_redone_original,
            ex_ws,
        ) = ContainerBlockLeafProcessor.__adjust_for_list_container_after_block_quote_special_special(
            parser_state, adj_original, bq_index, grab_bag
        )
        if (
            current_list_indent != -1
            and len(detabified_redone_original) >= current_list_indent
        ):
            return ex_ws
        return extracted_leaf_whitespace

    @staticmethod
    def __adjust_for_list_container_after_block_quote_kludge(
        parser_state: ParserState,
        xposition_marker: PositionMarker,
        calc_indent_level: int,
        extracted_leaf_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[str, str, Optional[str]]:
        removed_leading_space = None
        if not (
            "\t" not in grab_bag.original_line
            or grab_bag.text_removed_by_container != "> "
        ):
            POGGER.debug("__adjust_for_list_container_after_block_quote_kludge-->1")
            new_ex = ContainerBlockLeafProcessor.__adjust_for_list_container_after_block_quote_special(
                parser_state, xposition_marker, extracted_leaf_whitespace, grab_bag
            )
        else:
            POGGER.debug("__adjust_for_list_container_after_block_quote_kludge-->2")
            new_ex = extracted_leaf_whitespace
            POGGER.debug(
                "text_removed_by_container-->:$:<", grab_bag.text_removed_by_container
            )
            POGGER.debug("adj_line_to_parse-->:$:<", grab_bag.adj_line_to_parse)
            POGGER.debug("original_line-->:$:<", grab_bag.original_line)
            POGGER.debug(
                "text_removed_by_container-->:$:<", grab_bag.text_removed_by_container
            )
            if (
                grab_bag.text_removed_by_container
                and grab_bag.adj_line_to_parse
                and grab_bag.original_line
                and grab_bag.text_removed_by_container == "> "
            ):
                recon_line = (
                    grab_bag.text_removed_by_container + grab_bag.adj_line_to_parse
                )
                if (
                    recon_line == grab_bag.original_line
                    and extracted_leaf_whitespace
                    and len(extracted_leaf_whitespace) < calc_indent_level
                ):
                    extracted_leaf_whitespace = extracted_leaf_whitespace[
                        len(grab_bag.text_removed_by_container) :
                    ]
                    removed_leading_space = grab_bag.text_removed_by_container
        return new_ex, extracted_leaf_whitespace, removed_leading_space

    @staticmethod
    def __adjust_for_list_container_after_block_quote(
        parser_state: ParserState,
        xposition_marker: PositionMarker,
        last_list_index: int,
        extracted_leaf_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[
        Optional[ListStartMarkdownToken],
        Optional[str],
        Optional[str],
        Optional[ListStackToken],
        Optional[str],
    ]:
        list_token: Optional[ListStartMarkdownToken] = None
        POGGER.debug("yes adjust_for_list_container")
        removed_leading_space = None
        actual_removed_leading_space = None
        list_stack_token = None
        new_ex = None

        found_list_token = ContainerBlockLeafProcessor.__adjust_for_list_container_find(
            parser_state, xposition_marker
        )
        if not found_list_token:
            POGGER.debug(
                "__adjust_for_list_container_after_block_quote>>removed_leading_space>>$<<",
                removed_leading_space,
            )
            POGGER.debug(
                "__adjust_for_list_container_after_block_quote>>actual_removed_leading_space>>$<<",
                actual_removed_leading_space,
            )
            list_stack_token = cast(
                ListStackToken, parser_state.token_stack[last_list_index]
            )
            list_token = cast(
                ListStartMarkdownToken, list_stack_token.matching_markdown_token
            )
            calc_indent_level = (
                list_token.indent_level - len(grab_bag.text_removed_by_container)
                if grab_bag.text_removed_by_container
                else list_token.indent_level
            )
            if len(extracted_leaf_whitespace) > calc_indent_level:
                extracted_leaf_whitespace = extracted_leaf_whitespace[
                    :calc_indent_level
                ]

            (
                new_ex,
                extracted_leaf_whitespace,
                removed_leading_space,
            ) = ContainerBlockLeafProcessor.__adjust_for_list_container_after_block_quote_kludge(
                parser_state,
                xposition_marker,
                calc_indent_level,
                extracted_leaf_whitespace,
                grab_bag,
            )

            POGGER.debug(
                "__adjust_for_list_container_after_block_quote>>list_token>>$",
                list_token,
            )
            list_token.add_leading_spaces(new_ex)
            POGGER.debug(
                "__adjust_for_list_container_after_block_quote>>list_token>>$",
                list_token,
            )
            actual_removed_leading_space = extracted_leaf_whitespace

            if not grab_bag.container_depth and not xposition_marker.index_indent:
                removed_leading_space = extracted_leaf_whitespace
            POGGER.debug(
                "__adjust_for_list_container_after_block_quote>>removed_leading_space>>:$:<<",
                removed_leading_space,
            )
            POGGER.debug(
                "__adjust_for_list_container_after_block_quote>>actual_removed_leading_space>>:$:<<",
                actual_removed_leading_space,
            )
            POGGER.debug(
                "__adjust_for_list_container_after_block_quote>>new_ex>>:$:<<", new_ex
            )
        return (
            list_token,
            removed_leading_space,
            actual_removed_leading_space,
            list_stack_token,
            new_ex,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_list_container(
        parser_state: ParserState,
        xposition_marker: PositionMarker,
        last_block_index: int,
        last_list_index: int,
        extracted_leaf_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[Optional[str], Optional[str], PositionMarker, Optional[str]]:
        POGGER.debug("??? adjust_for_list_container")
        removed_leading_space = None
        actual_removed_leading_space = None
        list_token: Optional[ListStartMarkdownToken] = None
        new_ex = None
        list_stack_token = None
        # pylint: disable=chained-comparison
        if (
            not grab_bag.was_indent_already_processed
            and last_block_index > 0
            and last_list_index > 0
            and last_list_index > last_block_index
        ):
            (
                list_token,
                removed_leading_space,
                actual_removed_leading_space,
                list_stack_token,
                new_ex,
            ) = ContainerBlockLeafProcessor.__adjust_for_list_container_after_block_quote(
                parser_state,
                xposition_marker,
                last_list_index,
                extracted_leaf_whitespace,
                grab_bag,
            )
            if new_ex:
                POGGER.debug("new_ex>>:$:<<", new_ex)
                xposition_marker = PositionMarker(
                    xposition_marker.line_number,
                    len(new_ex),
                    xposition_marker.text_to_parse,
                    index_indent=xposition_marker.index_indent,
                )
                # POGGER.debug(
                #     "position_marker.text>>:$:<<", xposition_marker.text_to_parse
                # )
                # POGGER.debug(
                #     "position_marker.index>>:$:<<", xposition_marker.index_number
                # )
                # POGGER.debug(
                #     "position_marker.indent>>:$:<<", xposition_marker.index_indent
                # )
        else:
            POGGER.debug("not adjust_for_list_container")
        # pylint: enable=chained-comparison

        if ContainerBlockLeafProcessor.__adjust_for_list_container_kludge(
            parser_state,
            xposition_marker,
            removed_leading_space,
            actual_removed_leading_space,
            list_token,
            grab_bag,
            last_list_index,
            last_block_index,
        ):
            assert actual_removed_leading_space is not None
            assert (
                grab_bag.text_removed_by_container is not None
            ), "If here, some text must have been removed."
            total_removed = len(grab_bag.text_removed_by_container) + len(
                actual_removed_leading_space
            )
            assert list_stack_token is not None
            if total_removed < list_stack_token.start_index:
                delta = len(actual_removed_leading_space)
            else:
                assert list_token is not None
                delta = list_token.indent_level - total_removed
            xposition_marker = PositionMarker(
                xposition_marker.line_number,
                xposition_marker.index_number,
                xposition_marker.text_to_parse[delta:],
                xposition_marker.index_indent,
            )
            # POGGER.debug("position_marker.text>>:$:<<", xposition_marker.text_to_parse)
            # POGGER.debug("position_marker.index>>:$:<<", xposition_marker.index_number)
            # POGGER.debug("position_marker.indent>>:$:<<", xposition_marker.index_indent)
        return (
            removed_leading_space,
            actual_removed_leading_space,
            xposition_marker,
            new_ex,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-boolean-expressions,too-many-arguments
    @staticmethod
    def __adjust_for_list_container_kludge(
        parser_state: ParserState,
        xposition_marker: PositionMarker,
        removed_leading_space: Optional[str],
        actual_removed_leading_space: Optional[str],
        list_token: Optional[ListStartMarkdownToken],
        grab_bag: ContainerGrabBag,
        last_list_index: int,
        last_block_index: int,
    ) -> bool:
        apply_fix = False
        if (
            removed_leading_space is None
            and actual_removed_leading_space
            and list_token
            and grab_bag.block_quote_data.current_count
            == grab_bag.block_quote_data.stack_count
            and grab_bag.block_quote_data.stack_count > 1
            and grab_bag.is_para_continue
        ):
            keep_searching = True
            while keep_searching and last_list_index > last_block_index:
                inner_token = cast(
                    ListStackToken, parser_state.token_stack[last_list_index]
                )
                indent_delta = inner_token.indent_level - xposition_marker.index_indent
                line_to_parse = xposition_marker.text_to_parse
                if line_to_parse[:indent_delta].strip():
                    last_list_index -= 1
                else:
                    keep_searching = False

            if last_list_index == last_block_index:
                indent_delta = 0

            after_ws_index, extracted_whitespace = ParserHelper.extract_spaces_verified(
                line_to_parse, indent_delta
            )

            apply_fix = not LeafBlockProcessor.is_paragraph_ending_leaf_block_start(
                parser_state,
                line_to_parse,
                after_ws_index,
                extracted_whitespace,
                grab_bag.original_line,
                xposition_marker.index_indent,
            )
        return apply_fix

    # pylint: enable=too-many-boolean-expressions, too-many-arguments

    @staticmethod
    def __adjust_for_list_container_find(
        parser_state: ParserState, xposition_marker: PositionMarker
    ) -> Optional[MarkdownToken]:
        found_list_token, document_index = (
            None,
            len(parser_state.token_document) - 1,
        )
        if (
            document_index >= 0
            and parser_state.token_document[document_index].line_number
            == xposition_marker.line_number
        ):
            assert (
                parser_state.token_document[document_index].is_list_start
                or parser_state.token_document[document_index].is_new_list_item
            ), "Specified token is a list token."
            found_list_token = parser_state.token_document[document_index]
        return found_list_token

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_containers_before_leaf_blocks_adjust(
        parser_state: ParserState,
        xposition_marker: PositionMarker,
        last_block_index: int,
        total_ws: int,
        actual_removed_leading_space: Optional[str],
        grab_bag: ContainerGrabBag,
        new_ex: Optional[str],
    ) -> Tuple[str, int, Optional[PositionMarker]]:
        if grab_bag.do_force_list_continuation:
            POGGER.debug(
                "actual_removed_leading_space=$=", actual_removed_leading_space
            )
            if not actual_removed_leading_space:
                actual_removed_leading_space = ""
            assert xposition_marker.text_to_parse.startswith(
                actual_removed_leading_space
            ), "Text to parse must begin with the reported leading whitespace."
            position_marker = PositionMarker(
                xposition_marker.line_number,
                0,
                xposition_marker.text_to_parse[len(actual_removed_leading_space) :],
                xposition_marker.index_indent,
            )
            return "", -1, position_marker

        POGGER.debug("yes adjust_containers_before_leaf_blocks")
        (
            current_indent_level,
            close_tokens,
        ) = ContainerBlockLeafProcessor.__calculate_current_indent_level(
            parser_state,
            last_block_index,
            total_ws,
            xposition_marker,
            grab_bag,
        )
        POGGER.debug("close_tokens>:$:<", close_tokens)
        POGGER.debug("current_indent_level>:$:<", current_indent_level)
        POGGER.debug("xposition_marker.line_number>:$:<", xposition_marker.line_number)
        POGGER.debug(
            "xposition_marker.index_number>:$:<", xposition_marker.index_number
        )
        POGGER.debug(
            "xposition_marker.text_to_parse>:$:<", xposition_marker.text_to_parse
        )
        POGGER.debug(
            "xposition_marker.index_indent>:$:<", xposition_marker.index_indent
        )
        if close_tokens:
            grab_bag.extend_container_tokens(close_tokens)

        assert (
            parser_state.original_line_to_parse is not None
        ), "Original line must have been set by this point."
        (
            new_index_indent,
            new_text_to_parse,
        ) = ContainerBlockLeafProcessor.__make_adjustments(
            parser_state, xposition_marker, current_indent_level, grab_bag, new_ex
        )
        ContainerBlockLeafProcessor.__val(parser_state, new_text_to_parse)
        return new_text_to_parse, new_index_indent, None

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_containers_before_leaf_blocks(
        parser_state: ParserState,
        xposition_marker: PositionMarker,
        last_block_index: int,
        total_ws: int,
        actual_removed_leading_space: Optional[str],
        grab_bag: ContainerGrabBag,
        new_ex: Optional[str],
    ) -> PositionMarker:
        POGGER.debug("??? adjust_containers_before_leaf_blocks")
        POGGER.debug("xposition_marker.line_number>:$:<", xposition_marker.line_number)
        POGGER.debug(
            "xposition_marker.index_number>:$:<", xposition_marker.index_number
        )
        POGGER.debug(
            "xposition_marker.text_to_parse>:$:<", xposition_marker.text_to_parse
        )
        POGGER.debug(
            "xposition_marker.index_indent>:$:<", xposition_marker.index_indent
        )
        if (
            xposition_marker.text_to_parse
            and not grab_bag.was_paragraph_continuation
            and not grab_bag.was_indent_already_processed
        ):
            (
                new_text_to_parse,
                new_index_indent,
                new_position_marker,
            ) = ContainerBlockLeafProcessor.__adjust_containers_before_leaf_blocks_adjust(
                parser_state,
                xposition_marker,
                last_block_index,
                total_ws,
                actual_removed_leading_space,
                grab_bag,
                new_ex,
            )
            if new_position_marker:
                return new_position_marker
        else:
            POGGER.debug("not adjust_containers_before_leaf_blocks")
            new_text_to_parse = xposition_marker.text_to_parse
            new_index_indent = xposition_marker.index_indent

        indent_adjust = 0
        # if actual_removed_leading_space and new_index_indent == len(
        #     actual_removed_leading_space
        # ):
        #     # indent_adjust += new_index_indent
        #     # new_index_indent = 0
        #     pass

        POGGER.debug("xposition_marker.line_number>:$:<", xposition_marker.line_number)
        POGGER.debug("xposition_marker.index_number>:$:<", indent_adjust)
        POGGER.debug("xposition_marker.text_to_parse>:$:<", new_text_to_parse)
        POGGER.debug("xposition_marker.index_indent>:$:<", new_index_indent)
        return PositionMarker(
            xposition_marker.line_number,
            indent_adjust,
            new_text_to_parse,
            new_index_indent,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __calculate_current_indent_level_list(
        parser_state: ParserState, current_stack_index: int
    ) -> int:
        stack_token = cast(
            ListStackToken, parser_state.token_stack[current_stack_index]
        )
        if stack_token.last_new_list_token:
            return stack_token.last_new_list_token.indent_level
        list_token = cast(ListStartMarkdownToken, stack_token.matching_markdown_token)
        return list_token.indent_level

    @staticmethod
    def __post_leaf_block_adjustment(
        parser_state: ParserState,
        orig_text_removed_by_container: Optional[str],
        line_number: int,
    ) -> None:
        last_block_index = parser_state.find_last_block_quote_on_stack()
        POGGER.debug("last_block_index>>:$:", last_block_index)
        if not last_block_index:
            return

        last_block_token = cast(
            BlockQuoteMarkdownToken,
            parser_state.token_stack[last_block_index].matching_markdown_token,
        )

        if (
            not orig_text_removed_by_container
        ) and last_block_token.line_number != line_number:
            POGGER.debug(
                "plt-c>>last_block_token>>$",
                last_block_token,
            )
            POGGER.debug(
                "plt-c>>leading_text_index>>$", last_block_token.leading_text_index
            )
            POGGER.debug(
                "__post_leaf_block_adjustment>>block_token>>$", last_block_token
            )
            last_block_token.add_bleading_spaces("")
            POGGER.debug(
                "__post_leaf_block_adjustment>>block_token>>$", last_block_token
            )
            last_block_token.leading_text_index += 1
            POGGER.debug("plt-c>>last_block_token>>$", last_block_token)
            POGGER.debug(
                "plt-c>>leading_text_index>>$", last_block_token.leading_text_index
            )

    @staticmethod
    def __val(parser_state: ParserState, new_text_to_parse: str) -> None:
        assert (
            parser_state.original_line_to_parse is not None
        ), "Original line must have been set by this point."
        if len(parser_state.original_line_to_parse) == len(new_text_to_parse):
            assert parser_state.original_line_to_parse.replace(
                ">", ParserHelper.space_character
            ) == new_text_to_parse.replace(">", ParserHelper.space_character), (
                "cheat=:"
                + ParserHelper.make_value_visible(parser_state.original_line_to_parse)
                + ":,new_text_to_parse=:"
                + ParserHelper.make_value_visible(new_text_to_parse)
                + ":"
            )
        else:
            is_valid = parser_state.original_line_to_parse.endswith(new_text_to_parse)
            assert is_valid, (
                "cheat=:"
                + ParserHelper.make_value_visible(parser_state.original_line_to_parse)
                + ":,new_text_to_parse=:"
                + ParserHelper.make_value_visible(new_text_to_parse)
                + ":"
            )

    @staticmethod
    def __make_adjustments(
        parser_state: ParserState,
        xposition_marker: PositionMarker,
        current_indent_level: int,
        grab_bag: ContainerGrabBag,
        new_ex: Optional[str],
    ) -> Tuple[int, str]:
        assert (
            parser_state.original_line_to_parse is not None
        ), "Original line must have been set by this point."
        POGGER.debug(
            "parser_state.original_line_to_parse>>:$:($)",
            parser_state.original_line_to_parse,
            len(parser_state.original_line_to_parse),
        )
        POGGER.debug(
            "xposition_marker.text_to_parse>>:$:($)",
            xposition_marker.text_to_parse,
            len(xposition_marker.text_to_parse),
        )
        if grab_bag.weird_adjusted_text:
            new_index_indent = len(grab_bag.weird_adjusted_text)
            grab_bag.text_removed_by_container = grab_bag.weird_adjusted_text
            new_text_to_parse = parser_state.original_line_to_parse[new_index_indent:]
        else:
            POGGER.debug("current_indent_level>>:$:<", current_indent_level)
            current_indent_level -= xposition_marker.index_indent
            POGGER.debug("current_indent_level>>:$:<", current_indent_level)
            current_indent_level = max(current_indent_level, 0)
            assert current_indent_level >= 0, "Current indent must not go below 0."

            prefix_text = xposition_marker.text_to_parse[:current_indent_level]
            new_text_to_parse = xposition_marker.text_to_parse[current_indent_level:]
            new_index_indent = len(parser_state.original_line_to_parse) - len(
                new_text_to_parse
            )
            # if False and grab_bag.text_removed_by_container and ">" in grab_bag.text_removed_by_container:
            #     new_index_indent -= xposition_marker.index_indent

            POGGER.debug("new_ex>>:$:<", new_ex)
            POGGER.debug("new_text_to_parse>>:$:<", new_text_to_parse)
            POGGER.debug("new_index_indent>>:$:<", new_index_indent)
            POGGER.debug(
                "grab_bag.text_removed_by_container>>:$:<",
                grab_bag.text_removed_by_container,
            )
            POGGER.debug("prefix_text>>:$:<", prefix_text)
            grab_bag.text_removed_by_container = (
                grab_bag.text_removed_by_container + prefix_text
                if grab_bag.text_removed_by_container
                else prefix_text
            )
            new_index_indent, new_text_to_parse = (
                ContainerBlockLeafProcessor.__make_adjustments_inner(
                    parser_state, new_index_indent, new_text_to_parse, new_ex, grab_bag
                )
            )
            POGGER.debug(
                "grab_bag.text_removed_by_container>>:$:<",
                grab_bag.text_removed_by_container,
            )
        return new_index_indent, new_text_to_parse

    @staticmethod
    def __make_adjustments_inner(
        parser_state: ParserState,
        new_index_indent: int,
        new_text_to_parse: str,
        new_ex: Optional[str],
        grab_bag: ContainerGrabBag,
    ) -> Tuple[int, str]:
        adj_text_removed = grab_bag.text_removed_by_container
        assert adj_text_removed is not None
        if (
            len(adj_text_removed) > 1
            and adj_text_removed[-1] == " "
            and adj_text_removed[-2] == ">"
        ):
            adj_text_removed = adj_text_removed[2:]
        if (
            new_ex
            and not adj_text_removed.endswith(new_ex)
            and new_text_to_parse.startswith(new_ex)
        ):
            # assert new_ex is not None
            last_container_index = parser_state.find_last_container_on_stack()
            apply_adjustment = parser_state.token_stack[last_container_index].is_list
            if apply_adjustment and "\t" in grab_bag.original_line:
                apply_adjustment = (
                    ContainerBlockLeafProcessor.__make_adjustments_inner_tab_check(
                        grab_bag, new_ex
                    )
                )
            if apply_adjustment:
                new_index_indent += len(new_ex)
                new_text_to_parse = new_text_to_parse[len(new_ex) :]
        return new_index_indent, new_text_to_parse

    @staticmethod
    def __make_adjustments_inner_tab_check(
        grab_bag: ContainerGrabBag, new_ex: str
    ) -> bool:
        assert grab_bag.text_removed_by_container is not None
        removed_text = grab_bag.text_removed_by_container + new_ex
        index = 1
        line_part = None
        while True:
            assert index < len(grab_bag.original_line)
            line_part = grab_bag.original_line[:index]
            if len(TabHelper.detabify_string(line_part)) >= len(removed_text):
                break
            index += 1
        assert line_part is not None
        return "\t" not in line_part

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_current_indent_level_loop_kludge(
        parser_state: ParserState,
        total_ws: int,
        current_indent_level: int,
        last_list_index: int,
        current_stack_index: int,
        text_removed_by_container: Optional[str],
        non_last_block_index: int,
        last_block_index: int,
        line_number: int,
    ) -> Tuple[bool, int, Optional[int], int]:
        keep_processing = (
            total_ws != current_indent_level
            if (
                last_list_index > 0
                and text_removed_by_container is None
                and len(parser_state.token_stack) - 1 == current_stack_index
            )
            else True
        )
        if keep_processing:
            last_list_index = 0
            (
                new_indent_level,
                non_last_block_index,
            ) = ContainerBlockLeafProcessor.__calculate_current_indent_level_block_quote(
                parser_state,
                current_stack_index,
                non_last_block_index,
                last_block_index,
                line_number,
                current_indent_level,
                text_removed_by_container,
            )
            keep_processing = new_indent_level is not None
        else:
            new_indent_level = None

        return keep_processing, last_list_index, new_indent_level, non_last_block_index

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __xx_part_one(
        parser_state: ParserState,
        position_marker: PositionMarker,
        current_stack_index: int,
        total_ws: int,
        current_indent_level: int,
        last_list_index: int,
        text_removed_by_container: Optional[str],
        non_last_block_index: int,
        last_block_index: int,
    ) -> Tuple[Optional[int], bool, bool, int, int, bool]:
        proposed_indent_level: Optional[int] = 0
        POGGER.debug("token:$:", parser_state.token_stack[current_stack_index])
        continue_in_loop = True
        keep_processing = True
        had_non_block_token = False
        if parser_state.token_stack[current_stack_index].is_block_quote:
            (
                keep_processing,
                last_list_index,
                new_indent_level,
                non_last_block_index,
            ) = ContainerBlockLeafProcessor.__calculate_current_indent_level_loop_kludge(
                parser_state,
                total_ws,
                current_indent_level,
                last_list_index,
                current_stack_index,
                text_removed_by_container,
                non_last_block_index,
                last_block_index,
                position_marker.line_number,
            )
            if keep_processing:
                proposed_indent_level = new_indent_level
        elif parser_state.token_stack[current_stack_index].is_list:
            last_list_index = current_stack_index
            proposed_indent_level = (
                ContainerBlockLeafProcessor.__calculate_current_indent_level_list(
                    parser_state, current_stack_index
                )
            )
        else:
            had_non_block_token = True
            continue_in_loop = False
        return (
            proposed_indent_level,
            continue_in_loop,
            keep_processing,
            last_list_index,
            non_last_block_index,
            had_non_block_token,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=chained-comparison,too-many-arguments,too-many-locals
    @staticmethod
    def __xx_part_two(
        parser_state: ParserState,
        position_marker: PositionMarker,
        proposed_indent_level: Optional[int],
        total_ws: int,
        current_stack_index: int,
        current_indent_level: int,
        adj_ws: Optional[str],
        is_para_continue: bool,
        last_list_index: int,
        last_block_index: int,
        original_line: str,
        did_hit_indent_level_threshold: bool,
        continue_in_loop: bool,
    ) -> Tuple[bool, bool, int]:
        assert (
            proposed_indent_level is not None
        ), "Proposed ident level within lists must be decided."
        if proposed_indent_level > total_ws:
            did_hit_indent_level_threshold = True
            continue_in_loop = False
            if parser_state.token_stack[current_stack_index].is_list:
                is_total_ws_in_range = (
                    total_ws > current_indent_level and total_ws < proposed_indent_level
                )
                if (
                    is_total_ws_in_range
                    and not adj_ws
                    and is_para_continue
                    and last_list_index == last_block_index + 1
                ):
                    after_ws_index, extracted_whitespace = (
                        ParserHelper.extract_spaces_verified(
                            position_marker.text_to_parse,
                            position_marker.index_number,
                        )
                    )
                    keep_going = (
                        not LeafBlockProcessor.is_paragraph_ending_leaf_block_start(
                            parser_state,
                            position_marker.text_to_parse,
                            after_ws_index,
                            extracted_whitespace,
                            original_line,
                            position_marker.index_indent,
                        )
                    )
                    if keep_going:
                        current_indent_level += len(extracted_whitespace)
        else:
            current_indent_level = proposed_indent_level
            # POGGER.debug("current_indent_level:$", current_indent_level)
        return did_hit_indent_level_threshold, continue_in_loop, current_indent_level

    # pylint: enable=chained-comparison,too-many-arguments,too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __calculate_current_indent_level_loop(
        parser_state: ParserState,
        last_block_index: int,
        total_ws: int,
        position_marker: PositionMarker,
        current_stack_index: int,
        text_removed_by_container: Optional[str],
        current_indent_level: int,
        non_last_block_index: int,
        last_list_index: int,
        had_non_block_token: bool,
        did_hit_indent_level_threshold: bool,
        adj_ws: Optional[str],
        is_para_continue: bool,
        original_line: str,
    ) -> Tuple[int, int, int, bool, bool, bool]:
        (
            proposed_indent_level,
            continue_in_loop,
            keep_processing,
            last_list_index,
            non_last_block_index,
            had_non_block_token,
        ) = ContainerBlockLeafProcessor.__xx_part_one(
            parser_state,
            position_marker,
            current_stack_index,
            total_ws,
            current_indent_level,
            last_list_index,
            text_removed_by_container,
            non_last_block_index,
            last_block_index,
        )
        POGGER.debug(
            "proposed_indent_level:$ <= total_ws:$<",
            proposed_indent_level,
            total_ws,
        )
        if continue_in_loop and keep_processing:
            did_hit_indent_level_threshold, continue_in_loop, current_indent_level = (
                ContainerBlockLeafProcessor.__xx_part_two(
                    parser_state,
                    position_marker,
                    proposed_indent_level,
                    total_ws,
                    current_stack_index,
                    current_indent_level,
                    adj_ws,
                    is_para_continue,
                    last_list_index,
                    last_block_index,
                    original_line,
                    did_hit_indent_level_threshold,
                    continue_in_loop,
                )
            )
            POGGER.debug(
                "did_hit_indent_level_threshold>:$:<", did_hit_indent_level_threshold
            )
            POGGER.debug("continue_in_loop>:$:<", continue_in_loop)
            POGGER.debug("current_indent_level>:$:<", current_indent_level)
        return (
            current_indent_level,
            non_last_block_index,
            last_list_index,
            had_non_block_token,
            did_hit_indent_level_threshold,
            continue_in_loop,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __calculate_current_indent_level(
        parser_state: ParserState,
        last_block_index: int,
        total_ws: int,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[int, List[MarkdownToken]]:
        text_removed_by_container = grab_bag.text_removed_by_container
        current_indent_level = 0
        non_last_block_index = 0
        last_list_index = 0
        had_non_block_token = False
        did_hit_indent_level_threshold = False
        POGGER.debug("token-stack:$:", parser_state.token_stack)
        for current_stack_index in range(1, len(parser_state.token_stack)):
            POGGER.debug("<<current_indent_level:$", current_indent_level)
            (
                current_indent_level,
                non_last_block_index,
                last_list_index,
                had_non_block_token,
                did_hit_indent_level_threshold,
                continue_in_loop,
            ) = ContainerBlockLeafProcessor.__calculate_current_indent_level_loop(
                parser_state,
                last_block_index,
                total_ws,
                position_marker,
                current_stack_index,
                text_removed_by_container,
                current_indent_level,
                non_last_block_index,
                last_list_index,
                had_non_block_token,
                did_hit_indent_level_threshold,
                grab_bag.adj_ws,
                grab_bag.is_para_continue,
                grab_bag.original_line,
            )
            POGGER.debug("<<current_indent_level:$", current_indent_level)
            if not continue_in_loop:
                break
        POGGER.debug("<<current_indent_level:$", current_indent_level)
        close_tokens: List[MarkdownToken] = []
        if last_list_index:
            POGGER.debug("<<had_non_block_token:$", had_non_block_token)
            POGGER.debug("<<grab_bag.is_para_continue:$", grab_bag.is_para_continue)
            if (
                not had_non_block_token
                and did_hit_indent_level_threshold
                and not grab_bag.is_para_continue
            ):
                POGGER.debug("<<last_list_index:$", last_list_index)
                (
                    close_tokens,
                    grab_bag.requeue_line_info,
                ) = parser_state.close_open_blocks_fn(
                    parser_state,
                    include_lists=True,
                    was_forced=True,
                    until_this_index=last_list_index,
                )
                POGGER.debug("<<close_tokens:$", close_tokens)
        return current_indent_level, close_tokens

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_current_indent_level_block_quote(
        parser_state: ParserState,
        current_stack_index: int,
        non_last_block_index: int,
        last_block_index: int,
        line_number: int,
        current_indent_level: int,
        text_removed_by_container: Optional[str],
    ) -> Tuple[Optional[int], int]:
        if current_stack_index != last_block_index:
            POGGER.debug("not last bq token, skipping")
            non_last_block_index = current_stack_index
            return None, non_last_block_index

        matching_token = parser_state.token_stack[
            current_stack_index
        ].matching_markdown_token
        assert (
            matching_token is not None
        ), "Matching token is always set for containers."
        POGGER.debug("line_number=$", line_number)
        POGGER.debug(
            "matching_markdown_token.line_number=$", matching_token.line_number
        )
        POGGER.debug("non_last_block_index=$", non_last_block_index)

        valid_mark = non_last_block_index and non_last_block_index != (
            last_block_index - 1
        )
        POGGER.debug("current_indent_level=:$:", current_indent_level)
        base_indent_level = (
            current_indent_level
            if valid_mark and line_number == matching_token.line_number
            else 0
        )
        POGGER.debug("text_removed_by_container=:$:", text_removed_by_container)
        POGGER.debug("base_indent_level=:$:", base_indent_level)
        proposed_indent_level = (
            len(text_removed_by_container) if text_removed_by_container else 0
        ) + base_indent_level
        POGGER.debug("last bq token, processing:$", proposed_indent_level)
        return proposed_indent_level, non_last_block_index

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
Module to contain lists of the available extension token types.
"""

from typing import List

from pymarkdown.extensions.front_matter_markdown_token import FrontMatterMarkdownToken
from pymarkdown.extensions.pragma_token import PragmaToken
from pymarkdown.extensions.task_list_items import TaskListToken


# pylint: disable=too-few-public-methods
class ExtensionTokenTypes:
    """
    Class to contain lists of the available token types.
    """

    __TOKEN_TYPES: List[type] = [FrontMatterMarkdownToken, PragmaToken, TaskListToken]

    @staticmethod
    def get_token_types() -> List[type]:
        """
        Get a list of all available extension token types.
        """
        return ExtensionTokenTypes.__TOKEN_TYPES


# pylint: enable=too-few-public-methods



"""
Module to implement a plugin that looks for more than one space between either the
opening or closing hashes of an atx heading.
"""

from typing import Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd021(RulePlugin):
    """
    Class to implement a plugin that looks for more than one space between either the
    opening or closing hashes of an atx heading.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__atx_heading_token: Optional[AtxHeadingMarkdownToken] = None
        self.__is_left_in_error = False
        self.__first_text_token: Optional[TextMarkdownToken] = None
        self.__last_token: Optional[TextMarkdownToken] = None

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="no-multiple-space-closed-atx",
            plugin_id="MD021",
            plugin_enabled_by_default=True,
            plugin_description="Multiple spaces are present inside hash characters on Atx Closed Heading.",
            plugin_version="0.5.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md021.md",
            plugin_supports_fix=True,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__atx_heading_token = None
        self.__first_text_token = None
        self.__last_token = None
        self.__is_left_in_error = False

    def __report(
        self, context: PluginScanContext, token: MarkdownToken, extra_end_data: str
    ) -> None:
        assert self.__atx_heading_token is not None
        if context.in_fix_mode:
            if self.__is_left_in_error:
                assert self.__first_text_token is not None
                self.register_fix_token_request(
                    context,
                    self.__first_text_token,
                    "next_token",
                    "extracted_whitespace",
                    " ",
                )
            if len(extra_end_data) > 1:
                self.register_fix_token_request(
                    context,
                    token,
                    "next_token",
                    "extra_end_data",
                    " ",
                )
        else:
            self.report_next_token_error(context, self.__atx_heading_token)

    def __handle_atx_end(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        end_token = cast(EndMarkdownToken, token)
        assert end_token.extra_end_data is not None
        extra_end_data = end_token.extra_end_data
        if "\t" in extra_end_data:
            assert self.__last_token is not None
            assert self.__last_token.is_text
            resolved_token_text = ParserHelper.remove_all_from_text(
                self.__last_token.token_text
            )
            start_index = self.__last_token.column_number - 1
            resolved_token_text = TabHelper.detabify_string(
                resolved_token_text, start_index
            )
            start_index_after_last_text_token = start_index + len(resolved_token_text)
            extra_end_data = TabHelper.detabify_string(
                extra_end_data, start_index_after_last_text_token
            )
        if self.__is_left_in_error or len(extra_end_data) > 1:
            self.__report(context, token, extra_end_data)
        self.__atx_heading_token = None
        self.__first_text_token = None
        self.__last_token = None

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_atx_heading:
            atx_token = cast(AtxHeadingMarkdownToken, token)
            if atx_token.remove_trailing_count:
                self.__atx_heading_token = atx_token
            self.__is_left_in_error = False
        elif token.is_paragraph_end:
            self.__atx_heading_token = None
        elif token.is_atx_heading_end:
            self.__handle_atx_end(context, token)
        elif token.is_text:
            text_token = cast(TextMarkdownToken, token)
            if not self.__first_text_token and self.__atx_heading_token:
                resolved_extracted_whitespace = ParserHelper.remove_all_from_text(
                    text_token.extracted_whitespace
                )
                if "\t" in resolved_extracted_whitespace:
                    start_index = (
                        self.__atx_heading_token.column_number
                        - 1
                        + self.__atx_heading_token.hash_count
                    )
                    resolved_extracted_whitespace = TabHelper.detabify_string(
                        resolved_extracted_whitespace, start_index
                    )
                if len(resolved_extracted_whitespace) > 1:
                    self.__is_left_in_error = True
                self.__first_text_token = text_token
            self.__last_token = text_token



"""
Module to helper with determining whether to continue with the link reference definitions.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.links.link_parse_helper import LinkParseHelper
from pymarkdown.links.link_reference_definition_parse_helper import (
    LinkReferenceDefinitionParseHelper,
)
from pymarkdown.links.link_reference_tuple import LinkReferenceDefinitionTuple
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.link_reference_definition_markdown_token import (
    LinkReferenceDefinitionMarkdownToken,
)
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import LinkDefinitionStackToken

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-few-public-methods


class LinkReferenceDefinitionContinuationHelper:
    """
    Class to helper with determining whether to continue with the link reference definitions.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def determine_continue_or_stop(
        parser_state: ParserState,
        position_marker: PositionMarker,
        was_started: bool,
        remaining_line_to_parse: str,
        extracted_whitespace: str,
        unmodified_line_to_parse: str,
        original_stack_depth: int,
        original_document_depth: int,
        end_lrd_index: int,
        line_to_parse_size: int,
        is_blank_line: bool,
        did_complete_lrd: bool,
        parsed_lrd_tuple: Optional[LinkReferenceDefinitionTuple],
        lines_to_requeue: List[str],
    ) -> Tuple[bool, bool, List[MarkdownToken]]:
        """
        Determine whether to continue with the processing of the LRD.
        """
        if did_pause_lrd := (
            end_lrd_index >= 0
            and end_lrd_index == line_to_parse_size
            and not is_blank_line
        ):
            POGGER.debug(">>parse_link_reference_definition>>continuation")
            LinkReferenceDefinitionContinuationHelper.__add_line_for_lrd_continuation(
                parser_state,
                position_marker,
                was_started,
                remaining_line_to_parse,
                extracted_whitespace,
                unmodified_line_to_parse,
                original_stack_depth,
                original_document_depth,
            )
        if not did_pause_lrd and was_started or did_complete_lrd:
            return LinkReferenceDefinitionContinuationHelper.__stop_lrd_continuation(
                parser_state,
                did_complete_lrd,
                parsed_lrd_tuple,
                lines_to_requeue,
                did_pause_lrd,
            )

        POGGER.debug(">>parse_link_reference_definition>>other")
        return did_pause_lrd, False, []

    # pylint: enable=too-many-arguments

    @staticmethod
    def __stop_lrd_continuation(
        parser_state: ParserState,
        did_complete_lrd: bool,
        parsed_lrd_tuple: Optional[LinkReferenceDefinitionTuple],
        lines_to_requeue: List[str],
        did_pause_lrd: bool,
    ) -> Tuple[bool, bool, List[MarkdownToken]]:
        """
        As part of processing a link reference definition, stop a continuation.
        """
        POGGER.debug(">>parse_link_reference_definition>>no longer need start")
        if did_complete_lrd:
            assert parsed_lrd_tuple, "LRd tuple must be defined by now."
            assert (
                parsed_lrd_tuple.normalized_destination is not None
            ), "normalized_destination must be defined by now."
            did_add_definition = LinkParseHelper.add_link_definition(
                parsed_lrd_tuple.normalized_destination, parsed_lrd_tuple.link_titles
            )
            link_def_token = cast(
                LinkDefinitionStackToken, parser_state.token_stack[-1]
            )
            assert (
                link_def_token.extracted_whitespace is not None
            ), "extracted_whitespace must be defined by now."
            extracted_whitespace = link_def_token.extracted_whitespace

            POGGER.debug(
                "link_def_token.extracted_whitespace>:$:<",
                link_def_token.extracted_whitespace,
            )
            POGGER.debug(
                "link_def_token.continuation_lines>:$:<",
                link_def_token.continuation_lines,
            )
            POGGER.debug(
                "link_def_token.unmodified_lines>:$:<", link_def_token.unmodified_lines
            )
            POGGER.debug("lines_to_requeue>:$:<", lines_to_requeue)

            does_any_line_have_tabs = any(
                ParserHelper.tab_character in ffg
                for ffg in link_def_token.unmodified_lines
            )
            POGGER.debug("does_any_line_have_tabs>:$:<", does_any_line_have_tabs)

            last_container_index = parser_state.find_last_container_on_stack()
            if does_any_line_have_tabs and last_container_index > 0:
                (
                    extracted_whitespace,
                    parsed_lrd_tuple,
                ) = LinkReferenceDefinitionContinuationHelper.__stop_lrd_continuation_with_tab(
                    parser_state,
                    link_def_token,
                    parsed_lrd_tuple,
                )

            new_tokens: List[MarkdownToken] = (
                LinkReferenceDefinitionContinuationHelper.__create_new_tokens(
                    link_def_token,
                    parsed_lrd_tuple,
                    extracted_whitespace,
                    did_add_definition,
                )
            )
            del parser_state.token_stack[-1]
            return did_pause_lrd, len(lines_to_requeue) > 1, new_tokens

        del parser_state.token_stack[-1]
        return did_pause_lrd, True, []

    @staticmethod
    def __create_new_tokens(
        link_def_token: LinkDefinitionStackToken,
        parsed_lrd_tuple: LinkReferenceDefinitionTuple,
        extracted_whitespace: str,
        did_add_definition: bool,
    ) -> List[MarkdownToken]:
        assert (
            parsed_lrd_tuple.normalized_destination is not None
        ), "normalized_destination must be defined by now."
        new_tokens: List[MarkdownToken] = [
            LinkReferenceDefinitionMarkdownToken(
                did_add_definition,
                extracted_whitespace,
                parsed_lrd_tuple.normalized_destination,
                parsed_lrd_tuple.link_titles,
                parsed_lrd_tuple.link_info,
                position_marker=link_def_token.start_position_marker,
            )
        ]
        POGGER.debug(">>link_info>>$", parsed_lrd_tuple.link_info)
        POGGER.debug(">>new_tokens>>$", new_tokens)
        return new_tokens

    @staticmethod
    def __stop_lrd_continuation_with_tab(
        parser_state: ParserState,
        link_def_token: LinkDefinitionStackToken,
        parsed_lrd_tuple: LinkReferenceDefinitionTuple,
    ) -> Tuple[str, LinkReferenceDefinitionTuple]:
        POGGER.debug(
            "extracted_whitespace>:$:<",
            link_def_token.extracted_whitespace,
        )
        last_block_quote_index = parser_state.find_last_block_quote_on_stack()
        if last_block_quote_index:
            last_block_quote_token = parser_state.token_stack[last_block_quote_index]
            block_quote_token = cast(
                BlockQuoteMarkdownToken, last_block_quote_token.matching_markdown_token
            )
        else:
            block_quote_token = None
        last_list_index = parser_state.find_last_list_block_on_stack()
        if last_list_index:
            last_list_token = parser_state.token_stack[last_list_index]
            list_token = cast(
                ListStartMarkdownToken, last_list_token.matching_markdown_token
            )
        else:
            list_token = None

        POGGER.debug(
            "link_def_token.continuation_lines>:$:<", link_def_token.continuation_lines
        )
        POGGER.debug(
            "link_def_token.unmodified_lines>:$:<", link_def_token.unmodified_lines
        )

        if last_block_quote_index > last_list_index:
            assert block_quote_token is not None
            assert block_quote_token.bleading_spaces is not None
            split_container_spaces = block_quote_token.bleading_spaces.split("\n")
        else:
            assert list_token is not None
            assert list_token.leading_spaces is not None
            split_container_spaces = list_token.leading_spaces.split("\n")

        if len(link_def_token.continuation_lines) == 1:
            extracted_whitespace = LinkReferenceDefinitionContinuationHelper.__stop_lrd_continuation_with_tab_single(
                parser_state,
                link_def_token,
                last_block_quote_index,
                last_list_index,
                split_container_spaces,
            )
        else:
            (
                extracted_whitespace,
                parsed_lrd_tuple,
            ) = LinkReferenceDefinitionContinuationHelper.__stop_lrd_continuation_with_tab_multiple(
                parser_state,
                link_def_token,
                block_quote_token,
                last_block_quote_index,
                list_token,
                last_list_index,
                split_container_spaces,
            )

        return extracted_whitespace, parsed_lrd_tuple

    @staticmethod
    def __stop_lrd_continuation_with_tab_single(
        parser_state: ParserState,
        link_def_token: LinkDefinitionStackToken,
        last_block_quote_index: int,
        last_list_index: int,
        split_container_spaces: List[str],
    ) -> str:
        parsed_line = link_def_token.continuation_lines[0]
        original_line = link_def_token.unmodified_lines[0]

        current_line_container_spaces = (
            split_container_spaces[-2]
            if len(split_container_spaces) >= 2
            else split_container_spaces[-1]
        )
        (
            extracted_whitespace,
            split_tab,
            _,
        ) = LinkReferenceDefinitionContinuationHelper.__find_line_ws(
            parsed_line, original_line, current_line_container_spaces
        )

        if split_tab:
            if last_block_quote_index > last_list_index:
                TabHelper.adjust_block_quote_indent_for_tab(parser_state)
            else:
                TabHelper.adjust_block_quote_indent_for_tab(
                    parser_state, extracted_whitespace=extracted_whitespace
                )
        # POGGER.debug(
        #     "block_quote_token.leading_spaces>:$:<", block_quote_token.bleading_spaces
        # )
        return extracted_whitespace

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __stop_lrd_continuation_with_tab_multiple(
        parser_state: ParserState,
        link_def_token: LinkDefinitionStackToken,
        block_quote_token: Optional[BlockQuoteMarkdownToken],
        last_block_quote_index: int,
        list_token: Optional[ListStartMarkdownToken],
        last_list_index: int,
        split_container_spaces: List[str],
    ) -> Tuple[str, LinkReferenceDefinitionTuple]:
        split_tabs_list: List[bool] = []
        completed_lrd_text: str = ""
        extracted_whitespace = ""
        alt_ws: Optional[str] = None
        for this_line_index, this_line in enumerate(link_def_token.continuation_lines):
            (
                completed_lrd_text,
                extracted_whitespace,
                alt_ws,
            ) = LinkReferenceDefinitionContinuationHelper.__stop_lrd_continuation_with_tab_multiple_loop(
                link_def_token,
                this_line_index,
                this_line,
                completed_lrd_text,
                extracted_whitespace,
                alt_ws,
                split_tabs_list,
                split_container_spaces,
            )

        POGGER.debug("completed_lrd_text>:$:<", completed_lrd_text)
        assert alt_ws is not None, "This value must be set inside of the for loop."

        (
            did_succeed,
            next_index,
            new_parsed_lrd_tuple,
        ) = LinkReferenceDefinitionParseHelper.parse_link_reference_definition(
            parser_state, completed_lrd_text, 0, "", True, "", True
        )
        assert (
            did_succeed
        ), "Since this is the stop and there is at least one valid match, this must be true."
        assert (
            len(completed_lrd_text) == next_index
        ), "Index must be at the end of the stirng."
        assert new_parsed_lrd_tuple is not None, "New tuple must be defined."

        if last_block_quote_index > last_list_index:
            assert block_quote_token is not None
            LinkReferenceDefinitionContinuationHelper.__xx_multiple_fix_bleading_spaces(
                block_quote_token, split_tabs_list, link_def_token
            )
        else:
            assert list_token is not None
            LinkReferenceDefinitionContinuationHelper.__xx_multiple_fix_leading_spaces(
                list_token, split_tabs_list, link_def_token
            )
        return extracted_whitespace, new_parsed_lrd_tuple

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __stop_lrd_continuation_with_tab_multiple_loop(
        link_def_token: LinkDefinitionStackToken,
        this_line_index: int,
        this_line: str,
        completed_lrd_text: str,
        extracted_whitespace: str,
        alt_ws: Optional[str],
        split_tabs_list: List[bool],
        split_container_spaces: List[str],
    ) -> Tuple[str, str, Optional[str]]:
        original_this_line = link_def_token.unmodified_lines[this_line_index]
        POGGER.debug("this_line_index>:$:<", this_line_index)
        POGGER.debug("this_line>:$:<", this_line)
        POGGER.debug("original_this_line>:$:<", original_this_line)

        spaces_index = (
            len(split_container_spaces)
            - len(link_def_token.unmodified_lines)
            + this_line_index
        )
        (
            extracted_ws,
            split_tab,
            start_whitespace_index,
        ) = LinkReferenceDefinitionContinuationHelper.__find_line_ws(
            this_line, original_this_line, split_container_spaces[spaces_index]
        )

        if completed_lrd_text:
            completed_lrd_text += "\n"
        if this_line_index == 0:
            extracted_whitespace = extracted_ws
            alt_ws = TabHelper.detabify_string(
                extracted_whitespace, start_whitespace_index
            )
        else:
            completed_lrd_text += extracted_ws
        completed_lrd_text += this_line
        split_tabs_list.append(split_tab)
        return completed_lrd_text, extracted_whitespace, alt_ws

    # pylint: enable=too-many-arguments

    @staticmethod
    def __find_line_ws(
        parsed_lines: str, original_lines: str, wsx: str
    ) -> Tuple[str, bool, int]:
        start_text_index = original_lines.find(parsed_lines)
        assert start_text_index != -1, "Index must be found within string."
        # POGGER.debug("start_text_index>:$:<", start_text_index)
        start_whitespace_index = 0
        # POGGER.debug("start_whitespace_index>:$:<", start_whitespace_index)
        tabified_whitespace = original_lines[start_whitespace_index:start_text_index]
        # POGGER.debug("tabified_whitespace>:$:<", tabified_whitespace)

        split_tab = "\t" in tabified_whitespace
        if split_tab:
            tabified_whitespace_index = 1
            detabified_whitespace = ""
            detabified_length = -1
            while tabified_whitespace_index < (
                len(tabified_whitespace) + 1
            ) and detabified_length < len(wsx):
                detabified_whitespace = TabHelper.detabify_string(
                    tabified_whitespace[:tabified_whitespace_index]
                )
                detabified_length = len(detabified_whitespace)
                tabified_whitespace_index += 1
            tabified_whitespace_index -= 1
            does_end__with_bq_tab = (
                bool(wsx)
                and wsx[-1] == ">"
                and tabified_whitespace_index < len(tabified_whitespace)
                and tabified_whitespace[tabified_whitespace_index] == "\t"
            )
            split_tab = len(detabified_whitespace) != len(wsx) or does_end__with_bq_tab
        if split_tab:
            if not does_end__with_bq_tab:
                tabified_whitespace_index = 0
                while (
                    tabified_whitespace_index < len(wsx)
                    and wsx[tabified_whitespace_index] == " "
                    and tabified_whitespace[tabified_whitespace_index] == " "
                ):
                    tabified_whitespace_index += 1
            tabified_whitespace = tabified_whitespace[tabified_whitespace_index:]
        if not split_tab:
            tabified_whitespace = tabified_whitespace[len(wsx) :]
        extracted_whitespace = tabified_whitespace
        # POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        # POGGER.debug("split_tab>:$:<", split_tab)
        return extracted_whitespace, split_tab, start_whitespace_index

    @staticmethod
    def __xx_multiple_fix_bleading_spaces(
        block_quote_token: BlockQuoteMarkdownToken,
        split_tabs_list: List[bool],
        link_def_token: LinkDefinitionStackToken,
    ) -> None:
        POGGER.debug("split_tabs_list>:$:<", split_tabs_list)
        POGGER.debug(
            "block_quote_token.leading_spaces>:$:<", block_quote_token.bleading_spaces
        )
        assert (
            block_quote_token.bleading_spaces is not None
        ), "Bleading spaces must be defined by now."
        leading_spaces: List[str] = []
        for _ in link_def_token.continuation_lines:
            last_leading_space = block_quote_token.remove_last_bleading_space()
            POGGER.debug("last_leading_space>:$:<", last_leading_space)
            # if last_leading_space[0] == "\n":
            #     last_leading_space = last_leading_space[1:]
            leading_spaces.insert(0, last_leading_space)
        assert len(split_tabs_list) == len(
            leading_spaces
        ), "The two lists must have the same length."
        POGGER.debug("leading_spaces>:$:<", leading_spaces)
        POGGER.debug(
            "block_quote_token.leading_spaces>:$:<", block_quote_token.bleading_spaces
        )
        is_first = not block_quote_token.bleading_spaces
        for prefix_to_add in leading_spaces:
            # if split_tabs_list[0] and prefix_to_add[-1] == " "):
            #   prefix_to_add = prefix_to_add[:-1]
            assert not (split_tabs_list[0] and prefix_to_add[-1] == " ")
            del split_tabs_list[0]
            POGGER.debug(
                "__xx_multiple_fix_bleading_spaces>>block_token>>$", block_quote_token
            )
            block_quote_token.add_bleading_spaces(prefix_to_add, is_first)
            POGGER.debug(
                "__xx_multiple_fix_bleading_spaces>>block_token>>$", block_quote_token
            )
            is_first = False

    @staticmethod
    def __xx_multiple_fix_leading_spaces(
        list_token: ListStartMarkdownToken,
        split_tabs_list: List[bool],
        link_def_token: LinkDefinitionStackToken,
    ) -> None:
        POGGER.debug("split_tabs_list>:$:<", split_tabs_list)
        POGGER.debug("list_token.leading_spaces>:$:<", list_token.leading_spaces)
        assert (
            list_token.leading_spaces is not None
        ), "leading spaces must be defined by now."
        leading_spaces: List[str] = []
        for _ in link_def_token.continuation_lines:
            last_leading_space = list_token.remove_last_leading_space()
            POGGER.debug("last_leading_space>:$:<", last_leading_space)
            assert last_leading_space is not None
            leading_spaces.insert(0, last_leading_space)
        assert len(split_tabs_list) == len(
            leading_spaces
        ), "The two lists must have the same length."
        POGGER.debug("leading_spaces>:$:<", leading_spaces)
        POGGER.debug("list_token.leading_spaces>:$:<", list_token.leading_spaces)
        for current_leading_space_index, prefix_to_add in enumerate(leading_spaces):
            if split_tabs_list[0]:
                current_unmodified_line = link_def_token.unmodified_lines[
                    current_leading_space_index
                ]
                continuation_start_index = current_unmodified_line.find(
                    link_def_token.continuation_lines[current_leading_space_index]
                )
                unmodified_line_prefix = current_unmodified_line[
                    :continuation_start_index
                ]

                unmodified_line_prefix_index = 0
                detabified_length = -1
                while unmodified_line_prefix_index < len(
                    unmodified_line_prefix
                ) and detabified_length < len(prefix_to_add):
                    detabified_length = len(
                        TabHelper.detabify_string(
                            unmodified_line_prefix[: unmodified_line_prefix_index + 1]
                        )
                    )
                    unmodified_line_prefix_index += 1

                unmodified_line_prefix_index -= 1
                assert unmodified_line_prefix_index < len(
                    unmodified_line_prefix
                ), "Index must be within the string."
                assert unmodified_line_prefix[unmodified_line_prefix_index] == "\t"
                prefix_to_add = unmodified_line_prefix[:unmodified_line_prefix_index]
            del split_tabs_list[0]
            POGGER.debug(
                "__xx_multiple_fix_bleading_spaces>>block_token>>$", list_token
            )
            list_token.add_leading_spaces(prefix_to_add)
            POGGER.debug(
                "__xx_multiple_fix_bleading_spaces>>block_token>>$", list_token
            )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __add_line_for_lrd_continuation(
        parser_state: ParserState,
        position_marker: PositionMarker,
        was_started: bool,
        remaining_line_to_parse: str,
        extracted_whitespace: str,
        unmodified_line_to_parse: str,
        original_stack_depth: int,
        original_document_depth: int,
    ) -> None:
        """
        As part of processing a link reference definition, add a line to the continuation.
        """

        line_to_store = remaining_line_to_parse
        if not was_started:
            POGGER.debug(">>parse_link_reference_definition>>marking start")
            new_token = LinkDefinitionStackToken(extracted_whitespace, position_marker)
            parser_state.token_stack.append(new_token)
            new_token.original_stack_depth = original_stack_depth
            new_token.original_document_depth = original_document_depth

            new_token.last_block_quote_stack_token = (
                parser_state.last_block_quote_stack_token
            )
            new_token.last_block_quote_markdown_token_index = (
                parser_state.last_block_quote_markdown_token_index
            )
            new_token.copy_of_last_block_quote_markdown_token = (
                parser_state.copy_of_last_block_quote_markdown_token
            )
            new_token.copy_of_token_stack = parser_state.copy_of_token_stack

            new_token.x1_token = parser_state.x1_token
            new_token.copy_of_x1_token = parser_state.copy_of_x1_token
            new_token.x1_token_index = parser_state.x1_token_index
        else:
            new_token = cast(LinkDefinitionStackToken, parser_state.token_stack[-1])

        POGGER.debug(">>line_to_store>>add>:$<<", line_to_store)
        POGGER.debug(">>unmodified_line_to_parse>>add>:$<<", unmodified_line_to_parse)
        assert unmodified_line_to_parse.endswith(
            line_to_store
        ), "Unmodified line must end with the processed line."
        new_token.add_continuation_line(line_to_store)
        new_token.add_unmodified_line(unmodified_line_to_parse)

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
Module to provide for an encapsulation of the setext heading element.
"""

from typing import Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class SetextHeadingMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the setext heading element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        heading_character: str,
        heading_character_count: int,
        extracted_whitespace: str,
        position_marker: PositionMarker,
        para_token: ParagraphMarkdownToken,
    ) -> None:
        (
            self.__heading_character,
            self.__heading_character_count,
            self.__final_whitespace,
            self.__original_line_number,
            self.__original_column_number,
        ) = (
            heading_character,
            heading_character_count,
            "",
            para_token.line_number if para_token else -1,
            para_token.column_number if para_token else -1,
        )

        if self.__heading_character == "=":
            self.__hash_count = 1
        elif self.__heading_character == "-":
            self.__hash_count = 2
        else:
            self.__hash_count = -1

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_setext_heading,
            "",
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
            requires_end_token=True,
            can_force_close=False,
        )
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments
    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_setext_heading

    # pylint: enable=protected-access

    @property
    def final_whitespace(self) -> str:
        """
        Returns any final whitespace at the end of the heading that was removed.
        """
        return self.__final_whitespace

    @property
    def heading_character(self) -> str:
        """
        Returns the character associated with the heading start.
        """
        return self.__heading_character

    @property
    def hash_count(self) -> int:
        """
        Returns the count in equivalence of "Atx Hash" counts.
        """
        return self.__hash_count

    @property
    def heading_character_count(self) -> int:
        """
        Returns the count of characters associated with the heading start.
        """
        return self.__heading_character_count

    @property
    def original_line_number(self) -> int:
        """
        Returns the line number where this element actually started.
        """
        return self.__original_line_number

    @property
    def original_column_number(self) -> int:
        """
        Returns the column number where this element actually started.
        """
        return self.__original_column_number

    def set_final_whitespace(self, whitespace_to_set: str) -> None:
        """
        Set the final whitespace for the paragraph. That is any whitespace at the very
        end of the paragraph, removed to prevent hard lines at the end.
        """

        self.__final_whitespace = whitespace_to_set
        self.__compose_extra_data_field()

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """

        original_location = (
            f"({self.original_line_number},{self.original_column_number})"
        )
        field_parts = [
            self.__heading_character,
            str(self.__heading_character_count),
            self.extracted_whitespace,
            original_location,
        ]
        if self.final_whitespace:
            field_parts.append(self.final_whitespace)
        self._set_extra_data(MarkdownToken.extra_data_separator.join(field_parts))

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            SetextHeadingMarkdownToken,
            SetextHeadingMarkdownToken.__rehydrate_setext_heading,
            SetextHeadingMarkdownToken.__rehydrate_setext_heading_end,
        )

    @staticmethod
    def __rehydrate_setext_heading(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the setext heading from the token.
        """
        _ = previous_token

        context.block_stack.append(current_token)
        current_setext_token = cast(SetextHeadingMarkdownToken, current_token)
        return current_setext_token.extracted_whitespace

    @staticmethod
    def __rehydrate_setext_heading_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the end of the setext heading block from the token.
        """
        _ = (previous_token, next_token)

        current_start_token = cast(SetextHeadingMarkdownToken, context.block_stack[-1])
        current_end_token = cast(EndMarkdownToken, current_token)

        heading_character = current_start_token.heading_character
        heading_character_count = current_start_token.heading_character_count
        final_whitespace = current_start_token.final_whitespace
        del context.block_stack[-1]
        assert (
            current_end_token.extra_end_data is not None
        ), "extra_end_data must be defined by now"
        return "".join(
            [
                final_whitespace,
                ParserHelper.newline_character,
                current_end_token.extracted_whitespace,
                ParserHelper.repeat_string(heading_character, heading_character_count),
                current_end_token.extra_end_data,
                ParserHelper.newline_character,
            ]
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            SetextHeadingMarkdownToken,
            SetextHeadingMarkdownToken.__handle_start_setext_heading_token,
            SetextHeadingMarkdownToken.__handle_end_setext_heading_token,
        )

    @staticmethod
    def __handle_start_setext_heading_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = transform_state
        setext_token = cast(SetextHeadingMarkdownToken, next_token)

        token_parts = [output_html]
        if output_html.endswith("</ol>") or output_html.endswith("</ul>"):
            token_parts.append(ParserHelper.newline_character)
        token_parts.extend(
            ["<h", "1" if setext_token.heading_character == "=" else "2", ">"]
        )
        transform_state.is_in_setext_block = True
        return "".join(token_parts)

    @staticmethod
    def __handle_end_setext_heading_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        fenced_token_index = transform_state.actual_token_index - 1
        while not transform_state.actual_tokens[fenced_token_index].is_setext_heading:
            fenced_token_index -= 1
        fenced_token = cast(
            SetextHeadingMarkdownToken,
            transform_state.actual_tokens[fenced_token_index],
        )
        token_parts = [
            output_html,
            "</h",
            "1" if fenced_token.heading_character == "=" else "2",
            ">",
            ParserHelper.newline_character,
        ]
        transform_state.is_in_setext_block = False
        return "".join(token_parts)



"""
Module to provide helper functions for reintergrating tabs.
"""

import logging
from typing import Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_state import ParserState
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken

LOGGER = logging.getLogger(__name__)


class TabHelper:
    """
    Class to provide helper functions for reintergrating tabs.
    """

    @staticmethod
    def parse_thematic_break_with_tab(
        original_line: str, token_text: str, extracted_whitespace: str
    ) -> Tuple[str, bool, bool, Optional[str], str, int]:
        """
        Generic type of algorithm to deal with tabs, in this case, used by thematic breaks
        and HTML blocks.
        """

        LOGGER.debug(
            "original_line>>:%s:<",
            ParserHelper.make_whitespace_visible((original_line)),
        )
        LOGGER.debug("token_text>>:%s:<", ParserHelper.make_value_visible(token_text))
        LOGGER.debug(
            "extracted_whitespace>>:%s:<",
            ParserHelper.make_value_visible(extracted_whitespace),
        )
        (
            tabified_token_text,
            _,
            tabified_token_text_index,
        ) = TabHelper.find_detabify_string(
            original_line, token_text, use_proper_traverse=True
        )
        LOGGER.debug(
            "tabified_token_text>>:%s:<",
            ParserHelper.make_value_visible(tabified_token_text),
        )
        LOGGER.debug("tabified_token_text_index>>:%d:<", tabified_token_text_index)
        assert tabified_token_text_index != -1, "Detabified string must be found."
        assert tabified_token_text is not None, "Detabified string must be found."

        tabified_leading_spaces = original_line[:tabified_token_text_index]
        LOGGER.debug(
            "tabified_leading_spaces>>:%s:<",
            ParserHelper.make_value_visible(tabified_leading_spaces),
        )
        tabified_suffix = extracted_whitespace
        tabified_prefix = None

        split_tab = False
        split_tab_with_block_quote_suffix = False
        if tabified_leading_spaces:
            (
                tabified_prefix,
                tabified_suffix,
                split_tab,
                split_tab_with_block_quote_suffix,
            ) = TabHelper.match_tabbed_whitespace(
                extracted_whitespace, tabified_leading_spaces
            )
        LOGGER.debug(
            "tabified_token_text>>:%s:<",
            ParserHelper.make_value_visible(tabified_token_text),
        )
        LOGGER.debug(
            "tabified_prefix>>:%s:<", ParserHelper.make_value_visible(tabified_prefix)
        )
        LOGGER.debug(
            "tabified_suffix>>:%s:<", ParserHelper.make_value_visible(tabified_suffix)
        )
        LOGGER.debug("split_tab>>:%s:<", str(split_tab))
        LOGGER.debug(
            "split_tab_with_block_quote_suffix>>:%s:<",
            str(split_tab_with_block_quote_suffix),
        )
        return (
            tabified_token_text,
            split_tab,
            split_tab_with_block_quote_suffix,
            tabified_prefix,
            tabified_suffix,
            tabified_token_text_index,
        )

    @staticmethod
    def match_tabbed_whitespace(
        extracted_whitespace: str, corrected_extracted_whitespace: str
    ) -> Tuple[str, str, bool, bool]:
        """
        Match any tabbed whitespace with its non-tabbed counterpart.
        """
        detabified_suffix = ""
        detabified_prefix = ""
        corrected_prefix = ""
        corrected_suffix = corrected_extracted_whitespace
        index_from_end = len(corrected_extracted_whitespace) - 1
        have_been_inside_loop = False
        while index_from_end >= 0 and len(detabified_suffix) < len(
            extracted_whitespace
        ):
            have_been_inside_loop = True
            corrected_suffix = corrected_extracted_whitespace[index_from_end:]
            corrected_prefix = corrected_extracted_whitespace[:index_from_end]
            LOGGER.debug(
                "index_from_end=:%s: of :%s:",
                index_from_end,
                len(corrected_extracted_whitespace),
            )
            LOGGER.debug("corrected_suffix=:%s:", corrected_suffix)
            LOGGER.debug(
                "corrected_prefix=:%s:",
                ParserHelper.make_whitespace_visible(corrected_prefix),
            )
            detabified_prefix = TabHelper.detabify_string(corrected_prefix)
            LOGGER.debug("detabified_prefix=:%s:", detabified_prefix)
            detabified_suffix = TabHelper.detabify_string(
                corrected_suffix, additional_start_delta=len(detabified_prefix)
            )
            LOGGER.debug("detabified_suffix=:%s:", detabified_suffix)
            if len(detabified_suffix) < len(extracted_whitespace):
                index_from_end -= 1
        assert index_from_end >= 0, "Index must be within the string."

        if not have_been_inside_loop:
            corrected_prefix = corrected_extracted_whitespace
            corrected_suffix = ""

        LOGGER.debug(
            "corrected_prefix=:%s:", ParserHelper.make_value_visible(corrected_prefix)
        )
        LOGGER.debug(
            "corrected_suffix=:%s:", ParserHelper.make_value_visible(corrected_suffix)
        )
        split_tab = detabified_suffix != extracted_whitespace
        LOGGER.debug(
            "detabified_prefix=:%s:", ParserHelper.make_value_visible(detabified_prefix)
        )
        LOGGER.debug(
            "detabified_suffix=:%s:", ParserHelper.make_value_visible(detabified_suffix)
        )
        LOGGER.debug("split_tab=:%s:", str(split_tab))
        split_tab_with_block_quote_suffix = False
        if split_tab:
            split_tab_with_block_quote_suffix = detabified_prefix.endswith(">")
        elif detabified_prefix:
            assert detabified_prefix.endswith(
                " "
            ), "Detabified prefix must end with a space."
        #     assert detabified_suffix[1:] == extracted_whitespace

        return (
            corrected_prefix,
            corrected_suffix,
            split_tab,
            split_tab_with_block_quote_suffix,
        )

    @staticmethod
    def find_detabify_string_ex(
        original_line: str, detabified_line_to_match: str
    ) -> Tuple[str, int]:
        """
        Find a detabified line within the original line, automatically looking at all four tab offsets.
        """

        for initial_offset in range(4):
            (
                adjusted_original_line,
                original_index,
                _,
            ) = TabHelper.find_detabify_string(
                original_line, detabified_line_to_match, initial_offset
            )
            if adjusted_original_line is not None:
                return adjusted_original_line, original_index
        raise AssertionError("A detabbified string must be found.")

    @staticmethod
    def find_detabify_string_verified(
        original_line: str,
        detabified_line_to_match: str,
        initial_offset: int = 0,
        use_proper_traverse: bool = False,
    ) -> Tuple[str, int, int]:
        """
        Find a detabified line within the original line, asserting if the first return value is not defined.
        """
        adjusted_original_line, original_start_index, original_line_index = (
            TabHelper.find_detabify_string(
                original_line,
                detabified_line_to_match,
                initial_offset,
                use_proper_traverse,
            )
        )
        assert (
            adjusted_original_line is not None
        ), "adjusted_original_line must be defined."
        return adjusted_original_line, original_start_index, original_line_index

    @staticmethod
    def find_detabify_string(
        original_line: str,
        detabified_line_to_match: str,
        initial_offset: int = 0,
        use_proper_traverse: bool = False,
    ) -> Tuple[Optional[str], int, int]:
        """
        Find a detabified line within the original line.
        """

        # LOGGER.debug("original_line=:%s:", ParserHelper.make_whitespace_visible(original_line.replace("\t", "\\t")))
        # LOGGER.debug("detabified_line_to_match=:%s:", ParserHelper.make_whitespace_visible(detabified_line_to_match))
        # LOGGER.debug("initial_offset=:%d:", initial_offset)
        original_start_index = 0
        original_line_index = 0
        adjusted_original_line = original_line[original_line_index:]
        adjusted_start_index = original_start_index + initial_offset
        detabified_original_line = TabHelper.detabify_string(
            original_line, additional_start_delta=adjusted_start_index
        )
        # LOGGER.debug("detabified_original_line=:%s:", ParserHelper.make_whitespace_visible(str(detabified_original_line)))
        # LOGGER.debug("len(detabified_original_line)=%s >= len(detabified_line_to_match)=%s ", str(len(detabified_original_line)), str(len(detabified_line_to_match)))
        while (len(detabified_original_line)) >= len(detabified_line_to_match):
            adjusted_original_line = original_line[original_line_index:]
            adjusted_start_index = original_start_index + initial_offset
            # LOGGER.debug("adjusted_original_line=:%s:", adjusted_original_line.replace("\t", "\\t"))
            # LOGGER.debug("adjusted_start_index=:%d:", adjusted_start_index)
            detabified_original_line = TabHelper.detabify_string(
                adjusted_original_line,
                additional_start_delta=adjusted_start_index,
            )
            # LOGGER.debug("detabified_original_line=:%s:", ParserHelper.make_whitespace_visible(str(detabified_original_line)))
            if detabified_line_to_match == detabified_original_line:
                break
            if (
                use_proper_traverse
                and adjusted_original_line[0] == ParserHelper.tab_character
            ):
                original_start_index = (1 + (original_start_index // 4)) * 4
            else:
                original_start_index += 1
            original_line_index += 1
            # LOGGER.debug("len(detabified_original_line)=%s >= len(detabified_line_to_match)=%s ", str(len(detabified_original_line)), str(len(detabified_line_to_match)))
        if detabified_line_to_match == detabified_original_line:
            return adjusted_original_line, original_start_index, original_line_index
        return None, -1, -1

    @staticmethod
    def detabify_string(source_string: str, additional_start_delta: int = 0) -> str:
        """
        Given a string that may have one or more tabstops in it, resolve the
        tabstops into more easily handled space characters.
        """
        if ParserHelper.tab_character not in source_string:
            return source_string

        rebuilt_string = ""
        current_start_index = 0
        next_tab_index = source_string.find(ParserHelper.tab_character)
        # LOGGER.debug("next_tab_index=:%d:", next_tab_index)
        # LOGGER.debug("source_string=:%s:", ParserHelper.make_whitespace_visible(source_string.replace("\t", "\\t")))
        # LOGGER.debug("additional_start_delta=:%d:", additional_start_delta)
        while next_tab_index != -1:
            _, start_index = ParserHelper.collect_backwards_while_spaces_verified(
                source_string, next_tab_index
            )
            # LOGGER.debug("start_index=:%d:", start_index)
            end_index, _ = ParserHelper.collect_while_spaces(
                source_string, next_tab_index
            )
            # LOGGER.debug("end_index=:%d:", end_index)
            whitespace_section = source_string[start_index:end_index]
            # LOGGER.debug("whitespace_section=:%s:", ParserHelper.make_whitespace_visible(whitespace_section.replace("\t", "\\t")))
            if start_index:
                rebuilt_string += source_string[:start_index]
            realized_start_index = (
                current_start_index + start_index + additional_start_delta
            )
            # LOGGER.debug("realized_start_index=:%d:", realized_start_index)
            whitespace_actual_length = TabHelper.calculate_length(
                whitespace_section, realized_start_index
            )
            # LOGGER.debug("whitespace_actual_length=:%d:", whitespace_actual_length)
            rebuilt_string += ParserHelper.repeat_string(
                ParserHelper.space_character, whitespace_actual_length
            )
            current_start_index += start_index + whitespace_actual_length

            source_string = source_string[end_index:]
            next_tab_index = source_string.find(ParserHelper.tab_character)
            # LOGGER.debug("next_tab_index=:%d:", next_tab_index)
        if source_string:
            rebuilt_string += source_string
        return rebuilt_string

    @staticmethod
    def calculate_length(source_string: str, start_index: int = 0) -> int:
        """
        Calculate an adjusted length for the string.
        """

        string_length = start_index
        for source_character in source_string:
            string_length = (
                (int((string_length + 4) / 4) * 4)
                if source_character == ParserHelper.tab_character
                else (string_length + 1)
            )
        return string_length - start_index

    @staticmethod
    def is_length_less_than_or_equal_to(source_string: str, length_limit: int) -> bool:
        """
        Determine if the adjusted length of the string is less than or equal to the
        specified limit.
        """
        return TabHelper.calculate_length(source_string) <= length_limit

    @staticmethod
    def is_length_greater_than_or_equal_to(
        source_string: str, length_limit: int, start_index: int = 0
    ) -> bool:
        """
        Determine if the adjusted length of the string is greater than or equal to the
        specified limit.
        """
        return (
            TabHelper.calculate_length(source_string, start_index=start_index)
            >= length_limit
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def find_tabified_string(
        original_line: str,
        reconstructed_line: str,
        abc: bool = False,
        use_proper_traverse: bool = False,
        reconstruct_prefix: Optional[str] = None,
        was_indented: bool = False,
    ) -> Tuple[str, int, bool]:
        """
        Find the correct tabified string to represent the line, allowing
        for the possibility that the first character may be split inside
        of a tab character.
        """

        LOGGER.debug(
            "original_line>:%s:<", ParserHelper.make_value_visible(original_line)
        )
        LOGGER.debug(
            "reconstructed_line>:%s:<",
            ParserHelper.make_value_visible(reconstructed_line),
        )
        (
            adj_original,
            adj_original_index,
            adj_traverse_original_index,
        ) = TabHelper.find_detabify_string(
            original_line, reconstructed_line, use_proper_traverse=use_proper_traverse
        )
        LOGGER.debug(">>adj_original>:%s:<", adj_original)
        LOGGER.debug(">>adj_original_index>:%d:<", adj_original_index)
        LOGGER.debug(">>adj_traverse_original_index>:%d:<", adj_traverse_original_index)

        # This is a weird case.
        if adj_original is None and was_indented and original_line[0] == "\t":
            original_line_minus_leading_tab = original_line[1:]
            (
                adj_original,
                adj_original_index,
                adj_traverse_original_index,
            ) = TabHelper.find_detabify_string(
                original_line_minus_leading_tab,
                reconstructed_line,
                use_proper_traverse=use_proper_traverse,
            )
            was_indented = False
        split_tab = adj_original is None
        LOGGER.debug("split_tab>:%s:<", str(split_tab))
        if split_tab:
            (
                adj_original,
                adj_original_index,
                adj_traverse_original_index,
            ) = TabHelper.find_tabified_string_split(
                reconstructed_line,
                reconstruct_prefix,
                original_line,
                use_proper_traverse,
                abc,
            )
        assert (
            adj_original is not None
        ), "Adjusted original line must be defined by now."
        return_index = (
            adj_traverse_original_index if use_proper_traverse else adj_original_index
        )
        LOGGER.debug(
            "adj_original=:%s:",
            ParserHelper.make_whitespace_visible(adj_original.replace("\t", "\\t")),
        )
        LOGGER.debug(">>return_index>:%d:<", return_index)
        LOGGER.debug(">>split_tab>:%s:<", str(split_tab))
        return (
            adj_original,
            return_index,
            split_tab,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def find_tabified_string_split(
        reconstructed_line: str,
        reconstruct_prefix: Optional[str],
        original_line: str,
        use_proper_traverse: bool,
        abc: bool,
    ) -> Tuple[Optional[str], int, int]:
        """
        Find a tabified string within a split.
        """
        LOGGER.debug(
            ">>reconstructed_line>:%s:<",
            ParserHelper.make_whitespace_visible(
                reconstructed_line.replace("\t", "\\t")
            ),
        )
        # Need to split this tab between two areas.
        if not reconstruct_prefix:
            reconstruct_prefix = " "
        reconstructed_line = f"{reconstruct_prefix}{reconstructed_line}"
        LOGGER.debug(
            ">>reconstructed_line>:%s:<",
            ParserHelper.make_whitespace_visible(
                reconstructed_line.replace("\t", "\\t")
            ),
        )
        (
            adj_original,
            adj_original_index,
            adj_traverse_original_index,
        ) = TabHelper.find_detabify_string(
            original_line,
            reconstructed_line,
            use_proper_traverse=use_proper_traverse,
        )
        LOGGER.debug(">>adj_original>:%s:<", adj_original)
        LOGGER.debug(">>adj_original_index>:%d:<", adj_original_index)
        LOGGER.debug(">>adj_traverse_original_index>:%d:<", adj_traverse_original_index)
        if abc:
            adj_original_index += 1
            LOGGER.debug(">>adj_original_index>:%d:<", adj_original_index)
        return adj_original, adj_original_index, adj_traverse_original_index

    @staticmethod
    def __adjust_block_quote_indent_for_tab_block_quote(
        parser_state: ParserState, stack_token_index: int
    ) -> None:
        block_quote_token = cast(
            BlockQuoteMarkdownToken,
            parser_state.token_stack[stack_token_index].matching_markdown_token,
        )
        # POGGER.debug(
        #     "parser_state=:$:",
        #     block_quote_token,
        # )
        block_quote_leading_spaces = block_quote_token.bleading_spaces
        assert (
            block_quote_leading_spaces is not None
        ), "Block quote bleading spaces must be defined by now."
        # POGGER.debug("block_quote_leading_spaces=:$:", block_quote_leading_spaces)
        block_quote_leading_spaces_index = block_quote_leading_spaces.rfind("\n")
        last_block_quote_leading_space = block_quote_leading_spaces[
            block_quote_leading_spaces_index + 1 :
        ]
        # POGGER.debug(
        #     "last_block_quote_leading_space=:$:", last_block_quote_leading_space
        # )
        if last_block_quote_leading_space:
            assert last_block_quote_leading_space[-1] in [" ", ">"]
            if last_block_quote_leading_space[-1] != ">":
                last_block_quote_leading_space = last_block_quote_leading_space[:-1]
        # POGGER.debug(
        #     "last_block_quote_leading_space=:$:", last_block_quote_leading_space
        # )
        # POGGER.debug(
        #     "parser_state=:$:",
        #     block_quote_token,
        # )
        block_quote_token.remove_last_bleading_space()
        # POGGER.debug(
        #     "parser_state=:$:",
        #     block_quote_token,
        # )

        LOGGER.debug(
            "__adjust_block_quote_indent_for_tab_block_quote>>block_token>>%s",
            str(block_quote_token),
        )
        block_quote_token.add_bleading_spaces(last_block_quote_leading_space)
        LOGGER.debug(
            "__adjust_block_quote_indent_for_tab_block_quote>>block_token>>%s",
            str(block_quote_token),
        )
        # POGGER.debug(
        #     "parser_state=:$:",
        #     block_quote_token,
        # )

    @staticmethod
    def __adjust_block_quote_indent_for_tab_list_normal(
        last_list_leading_space: str,
        fenced_switch_enabled: bool,
        tab_index: int,
        alternate_list_leading_space: Optional[str],
        extracted_whitespace: str,
    ) -> Tuple[str, str]:
        last_list_leading_space_length = len(last_list_leading_space)
        if fenced_switch_enabled:
            assert tab_index < last_list_leading_space_length or (
                last_list_leading_space_length == 0
            ), "Tab index must be within leading spaces."  # and tab_index == 0)
        else:
            assert (
                tab_index < last_list_leading_space_length
            ), "Tab index must be within leading spaces."
        last_list_leading_space = (
            extracted_whitespace[:tab_index]
            if alternate_list_leading_space is None
            else alternate_list_leading_space
        )
        extracted_whitespace = extracted_whitespace[tab_index:]
        return last_list_leading_space, extracted_whitespace

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_block_quote_indent_for_tab_list(
        parser_state: ParserState,
        stack_token_index: int,
        extracted_whitespace: str,
        fenced_switch_enabled: bool,
        alternate_list_leading_space: Optional[str],
        original_line: Optional[str] = None,
    ) -> str:
        list_start_token = cast(
            ListStartMarkdownToken,
            parser_state.token_stack[stack_token_index].matching_markdown_token,
        )
        LOGGER.debug(
            "list_start_token=:%s:",
            ParserHelper.make_value_visible(list_start_token),
        )
        list_leading_spaces = list_start_token.leading_spaces
        assert list_leading_spaces is not None, "Leading spaces must be defined by now."
        LOGGER.debug(
            "list_leading_spaces=:%s:",
            ParserHelper.make_value_visible(list_leading_spaces),
        )
        list_leading_spaces_index = list_leading_spaces.rfind("\n")
        last_list_leading_space = list_leading_spaces[list_leading_spaces_index + 1 :]
        LOGGER.debug(
            "last_list_leading_space=:%s:",
            ParserHelper.make_value_visible(last_list_leading_space),
        )
        tab_index = extracted_whitespace.find("\t")
        LOGGER.debug("extracted_whitespace=:%s:", extracted_whitespace)
        LOGGER.debug("tab_index=:%d:", tab_index)

        if (
            tab_index == -1
            and extracted_whitespace
            and last_list_leading_space
            and last_list_leading_space != extracted_whitespace
            and original_line is not None
        ):
            last_list_leading_space = (
                TabHelper.__adjust_block_quote_indent_for_tab_list_kludge(
                    original_line, extracted_whitespace, last_list_leading_space
                )
            )
        else:
            (
                last_list_leading_space,
                extracted_whitespace,
            ) = TabHelper.__adjust_block_quote_indent_for_tab_list_normal(
                last_list_leading_space,
                fenced_switch_enabled,
                tab_index,
                alternate_list_leading_space,
                extracted_whitespace,
            )
        LOGGER.debug("last_list_leading_space=:%s:", last_list_leading_space)
        LOGGER.debug("extracted_whitespace=:%s:", extracted_whitespace)

        LOGGER.debug(
            "list_start_token=:%s:",
            ParserHelper.make_value_visible(list_start_token),
        )
        list_start_token.remove_last_leading_space()
        LOGGER.debug(
            "list_start_token=:%s:",
            ParserHelper.make_value_visible(list_start_token),
        )
        list_start_token.add_leading_spaces(last_list_leading_space)
        LOGGER.debug(
            "list_start_token=:%s:",
            ParserHelper.make_value_visible(list_start_token),
        )
        return extracted_whitespace

    # pylint: enable=too-many-arguments
    @staticmethod
    def adjust_block_quote_indent_for_tab_verified(
        parser_state: ParserState,
        extracted_whitespace: Optional[str] = None,
        alternate_list_leading_space: Optional[str] = None,
        fenced_switch_enabled: bool = False,
        original_line: Optional[str] = None,
    ) -> str:
        """
        Adjust the last block quote for a tab, asserting if the returned string is not defined.
        """
        extracted_whitespace = TabHelper.adjust_block_quote_indent_for_tab(
            parser_state,
            extracted_whitespace,
            alternate_list_leading_space,
            fenced_switch_enabled,
            original_line,
        )
        assert (
            extracted_whitespace is not None
        ), "extracted_whitespace must be defined by now."
        return extracted_whitespace

    @staticmethod
    def adjust_block_quote_indent_for_tab(
        parser_state: ParserState,
        extracted_whitespace: Optional[str] = None,
        alternate_list_leading_space: Optional[str] = None,
        fenced_switch_enabled: bool = False,
        original_line: Optional[str] = None,
    ) -> Optional[str]:
        """
        Adjust the last block quote for a tab.
        """
        LOGGER.debug(
            "extracted_whitespace=:%s:",
            ParserHelper.make_value_visible(extracted_whitespace),
        )
        LOGGER.debug(
            "parser_state=:%s:",
            ParserHelper.make_value_visible(parser_state.token_stack),
        )
        stack_token_index = len(parser_state.token_stack) - 1
        while (
            stack_token_index > 0
            and not parser_state.token_stack[stack_token_index].is_block_quote
            and not parser_state.token_stack[stack_token_index].is_list
        ):
            stack_token_index -= 1
        assert stack_token_index != 0, "Must not have gone back to the root index."

        LOGGER.debug(
            "parser_state=:%s:",
            ParserHelper.make_value_visible(
                parser_state.token_stack[stack_token_index]
            ),
        )
        if parser_state.token_stack[stack_token_index].is_block_quote:
            TabHelper.__adjust_block_quote_indent_for_tab_block_quote(
                parser_state, stack_token_index
            )
        else:
            assert (
                extracted_whitespace is not None
            ), "If we are not looking at a block quote, extracted_whitespace must be defined."
            extracted_whitespace = TabHelper.__adjust_block_quote_indent_for_tab_list(
                parser_state,
                stack_token_index,
                extracted_whitespace,
                fenced_switch_enabled,
                alternate_list_leading_space,
                original_line,
            )
        return extracted_whitespace

    @staticmethod
    def __adjust_block_quote_indent_for_tab_list_kludge(
        original_line: str, extracted_whitespace: str, last_list_leading_space: str
    ) -> str:
        orig_end_ws_index, _ = ParserHelper.extract_ascii_whitespace(original_line, 0)
        orig_after_ws = original_line[orig_end_ws_index:]
        recon_line = last_list_leading_space + extracted_whitespace[:-1] + orig_after_ws

        reconstructed_line, _, _ = TabHelper.find_tabified_string_split(
            recon_line, "", original_line, True, False
        )
        assert reconstructed_line is not None, "reconstructed_line must be found."
        text_to_use_count = 1
        keep_going = text_to_use_count < len(reconstructed_line)
        while keep_going:
            detabbified_line = TabHelper.detabify_string(
                reconstructed_line[:text_to_use_count]
            )
            keep_going = text_to_use_count < len(reconstructed_line) and len(
                detabbified_line
            ) < len(last_list_leading_space)
            if keep_going:
                text_to_use_count += 1
        assert (
            reconstructed_line[text_to_use_count - 1] == "\t"
        ), "Last character must be a tab."
        return reconstructed_line[: text_to_use_count - 1]

    @staticmethod
    def search_for_tabbed_prefix(
        extracted_space: str, detabify_index: int, lead_space_len: int
    ) -> Tuple[int, str, str]:
        """
        Look for a specific tabbed prefix length within a given string.
        """
        extra_detabify_index = 0
        indent_used = 0

        search_index = 1
        keep_going = True
        detabbed_ex_part = ex_part = ""
        while keep_going:
            ex_part = extracted_space[indent_used : indent_used + search_index]
            detabbed_ex_part = TabHelper.detabify_string(
                ex_part, detabify_index + extra_detabify_index
            )
            keep_going = (
                search_index < len(extracted_space) + 1
                and len(detabbed_ex_part) < lead_space_len
            )
            if keep_going:
                search_index += 1
        assert (
            len(detabbed_ex_part) >= lead_space_len
        ), "Proper amount of space must be used up."
        return search_index, ex_part, detabbed_ex_part

    @staticmethod
    def search_for_tabbed_prefix2(
        extracted_space: str,
        detabify_index: int,
        extra_detabify_index: int,
        indent_used: int,
        lead_space_len: int,
    ) -> Tuple[int, str, str]:
        """
        Look for a specific tabbed prefix length within a given string.
        """
        search_index = 1
        keep_going = True
        ex_part = detabbed_ex_part = ""
        while keep_going:
            ex_part = extracted_space[indent_used : indent_used + search_index]
            detabbed_ex_part = TabHelper.detabify_string(
                ex_part, detabify_index + extra_detabify_index
            )
            keep_going = (
                len(detabbed_ex_part) < len(extracted_space)
                and len(detabbed_ex_part) < lead_space_len
            )
            if keep_going:
                search_index += 1
        assert (
            len(detabbed_ex_part) >= lead_space_len
        ), "Proper amount of space must be used up."
        return search_index, ex_part, detabbed_ex_part



"""
Module to provide for an encapsulation of the link element.
"""

import logging
from typing import List, Optional, cast

from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.links.link_helper_properties import LinkHelperProperties
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.reference_markdown_token import ReferenceMarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class LinkStartMarkdownToken(ReferenceMarkdownToken):
    """
    Class to provide for an encapsulation of the link element.
    """

    def __init__(
        self,
        text_from_blocks: str,
        line_number: int,
        column_number: int,
        lhp: LinkHelperProperties,
    ) -> None:
        ReferenceMarkdownToken.__init__(
            self,
            lhp,
            LinkStartMarkdownToken.get_markdown_token_type(),
            "",
            text_from_blocks,
            line_number=line_number,
            column_number=column_number,
            requires_end_token=True,
            can_force_close=False,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_inline_link

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            LinkStartMarkdownToken,
            LinkStartMarkdownToken.__rehydrate_inline_link,
            LinkStartMarkdownToken.__rehydrate_inline_link_end,
        )

    @staticmethod
    def __rehydrate_inline_link(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the start of the link from the token.
        """
        _ = previous_token

        inline_current_token = cast(LinkStartMarkdownToken, current_token)
        context.block_stack.append(current_token)
        rehydrated_text = LinkStartMarkdownToken.rehydrate_inline_link_text_from_token(
            inline_current_token
        )
        return LinkStartMarkdownToken.insert_leading_whitespace_at_newlines(
            context, rehydrated_text
        )

    @staticmethod
    def __rehydrate_inline_link_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the end of the link from the token.
        """
        _ = (current_token, previous_token, next_token)

        del context.block_stack[-1]
        return ""

    @staticmethod
    def rehydrate_inline_link_text_from_token(
        link_token: ReferenceMarkdownToken,
    ) -> str:
        """
        Given a link token, rehydrate it's original text from the token.
        """

        link_parts = []
        if link_token.label_type == Constants.link_type__shortcut:
            link_parts.extend(
                [
                    "[",
                    ParserHelper.remove_all_from_text(link_token.text_from_blocks),
                    "]",
                ]
            )
        elif link_token.label_type == Constants.link_type__full:
            assert link_token.ex_label is not None, "Full links must have a label"
            link_parts.extend(
                ["[", link_token.text_from_blocks, "][", link_token.ex_label, "]"]
            )
        elif link_token.label_type == Constants.link_type__collapsed:
            link_parts.extend(["[", link_token.text_from_blocks, "][]"])
        else:
            assert (
                link_token.label_type == Constants.link_type__inline
            ), "The default is an inline link."
            LinkStartMarkdownToken.__rehydrate_inline_link_text_from_token_type_inline(
                link_token, link_parts
            )

        return "".join(link_parts)

    @staticmethod
    def __rehydrate_inline_link_text_from_token_type_inline(
        link_token: ReferenceMarkdownToken, link_parts: List[str]
    ) -> None:
        assert (
            link_token.before_title_whitespace is not None
        ), "Before title whitespace must be defined."
        assert (
            link_token.before_link_whitespace is not None
        ), "Before link whitespace must be defined."
        link_parts.extend(
            [
                "[",
                ParserHelper.remove_all_from_text(link_token.text_from_blocks),
                "](",
                link_token.before_link_whitespace,
                (
                    f"<{link_token.active_link_uri}>"
                    if link_token.did_use_angle_start
                    else link_token.active_link_uri
                ),
                link_token.before_title_whitespace,
            ]
        )
        if link_token.active_link_title:
            if link_token.inline_title_bounding_character == "'":
                title_prefix = "'"
                title_suffix = "'"
            elif link_token.inline_title_bounding_character == "(":
                title_prefix = "("
                title_suffix = ")"
            else:
                title_prefix = '"'
                title_suffix = '"'

            assert (
                link_token.after_title_whitespace is not None
            ), "After title whitespace must be defined."
            link_parts.extend(
                [
                    title_prefix,
                    link_token.active_link_title,
                    title_suffix,
                    link_token.after_title_whitespace,
                ]
            )
        link_parts.append(")")

    @staticmethod
    def insert_leading_whitespace_at_newlines(
        context: MarkdownTransformContext, text_to_modify: str
    ) -> str:
        """
        Deal with re-inserting any removed whitespace at the starts of lines.
        """
        if ParserHelper.newline_character in text_to_modify:
            owning_paragraph_token = next(
                (
                    context.block_stack[search_index]
                    for search_index in range(len(context.block_stack) - 1, -1, -1)
                    if context.block_stack[search_index].is_paragraph
                ),
                None,
            )

            POGGER.debug(
                f"text>before>{ParserHelper.make_value_visible(text_to_modify)}"
            )
            text_to_modify = ParserHelper.remove_all_from_text(text_to_modify)
            POGGER.debug(
                f"text>after>{ParserHelper.make_value_visible(text_to_modify)}"
            )

            if owning_paragraph_token:
                paragraph_token = cast(ParagraphMarkdownToken, owning_paragraph_token)
                (
                    text_to_modify,
                    paragraph_token.rehydrate_index,
                ) = ParserHelper.recombine_string_with_whitespace(
                    text_to_modify,
                    paragraph_token.extracted_whitespace,
                    paragraph_token.rehydrate_index,
                    post_increment_index=False,
                )
        return text_to_modify

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            LinkStartMarkdownToken,
            LinkStartMarkdownToken.__handle_start_link_token,
            LinkStartMarkdownToken.__handle_end_link_token,
        )

    @staticmethod
    def __handle_start_link_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = transform_state

        link_token = cast(LinkStartMarkdownToken, next_token)
        return "".join(
            [
                output_html,
                '<a href="',
                link_token.link_uri,
                f'" title="{link_token.link_title}' if link_token.link_title else "",
                '">',
            ]
        )

    @staticmethod
    def __handle_end_link_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (next_token, transform_state)

        return f"{output_html}</a>"



"""
Module to implement a plugin that looks for text in a paragraph where a line starts
with what could be an atx heading, except there is no spaces between the hashes and
the text of the heading.
"""

import re
from typing import Any, Optional, Tuple, cast

from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.link_start_markdown_token import LinkStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class StartOfLineTokenParser:
    """
    Token parser that activates a check only on valid newlines within
    a paragraph.
    """

    def __init__(self) -> None:
        self.__last_paragraph_token: Optional[ParagraphMarkdownToken] = None
        self.__paragraph_index = -1
        self.__first_line_after_other_token = False
        self.__paragraph_column_number = 0
        self.__inside_of_link = False
        self.__first_line_after_hard_break = False
        self.__delayed_line: Optional[
            Tuple[str, PluginScanContext, MarkdownToken, int, int]
        ] = None

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__last_paragraph_token = None
        self.__paragraph_index = -1
        self.__first_line_after_other_token = False
        self.__paragraph_column_number = 0

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_paragraph:
            paragraph_token = cast(ParagraphMarkdownToken, token)
            self.__next_token_paragraph_start(paragraph_token)
        elif token.is_paragraph_end:
            self.__next_token_paragraph_end()
        elif self.__last_paragraph_token:
            if self.__inside_of_link:
                if token.is_inline_link_end:
                    self.__inside_of_link = False
            elif token.is_text:
                text_token = cast(TextMarkdownToken, token)
                self.__next_token_paragraph_text_inline(text_token, context)
            else:
                self.__next_token_paragraph_non_text_inline(token)

    def __next_token_paragraph_start(self, token: ParagraphMarkdownToken) -> None:
        self.__last_paragraph_token = token
        self.__paragraph_index = 0
        self.__first_line_after_other_token = True
        self.__first_line_after_hard_break = False
        self.__inside_of_link = False
        self.__paragraph_column_number = token.column_number
        self.__delayed_line = None

    def __next_token_paragraph_end(self) -> None:
        if self.__delayed_line:
            self.check_start_of_line(
                self.__delayed_line[0],
                self.__delayed_line[1],
                self.__delayed_line[2],
                self.__delayed_line[3],
                self.__delayed_line[4],
            )
            self.__delayed_line = None
        self.__last_paragraph_token = None

    def __next_token_paragraph_non_text_inline_image(
        self, token: MarkdownToken
    ) -> None:  # sourcery skip: extract-method
        link_token = cast(LinkStartMarkdownToken, token)
        self.__paragraph_index += link_token.text_from_blocks.count(
            ParserHelper.newline_character
        )
        if link_token.label_type == Constants.link_type__inline:
            assert link_token.before_link_whitespace is not None
            self.__paragraph_index += link_token.before_link_whitespace.count(
                ParserHelper.newline_character
            )
            assert link_token.before_title_whitespace is not None
            self.__paragraph_index += link_token.before_title_whitespace.count(
                ParserHelper.newline_character
            )
            assert link_token.after_title_whitespace is not None
            self.__paragraph_index += link_token.after_title_whitespace.count(
                ParserHelper.newline_character
            )
            assert link_token.active_link_title is not None
            self.__paragraph_index += link_token.active_link_title.count(
                ParserHelper.newline_character
            )
        if link_token.label_type == Constants.link_type__full:
            assert link_token.ex_label is not None
            self.__paragraph_index += link_token.ex_label.count(
                ParserHelper.newline_character
            )
        self.__inside_of_link = token.is_inline_link

    def __next_token_paragraph_non_text_inline(self, token: MarkdownToken) -> None:
        if token.is_inline_code_span:
            code_span_token = cast(InlineCodeSpanMarkdownToken, token)
            self.__paragraph_index += (
                code_span_token.leading_whitespace.count(ParserHelper.newline_character)
                + code_span_token.span_text.count(ParserHelper.newline_character)
                + code_span_token.trailing_whitespace.count(
                    ParserHelper.newline_character
                )
            )
        elif token.is_inline_raw_html:
            raw_html_token = cast(RawHtmlMarkdownToken, token)
            self.__paragraph_index += raw_html_token.raw_tag.count(
                ParserHelper.newline_character
            )
        elif token.is_inline_image or token.is_inline_link:
            self.__next_token_paragraph_non_text_inline_image(token)
        elif token.is_inline_hard_break:
            self.__paragraph_index += 1
            self.__first_line_after_hard_break = True
        else:
            assert (
                token.is_inline_emphasis
                or token.is_inline_emphasis_end
                or token.is_inline_autolink
                or token.is_task_list
            )
        self.__delayed_line = None

    def __next_token_paragraph_text_inline(
        self, token: TextMarkdownToken, context: PluginScanContext
    ) -> None:
        assert self.__last_paragraph_token is not None
        split_whitespace = self.__last_paragraph_token.extracted_whitespace.split(
            ParserHelper.newline_character
        )
        split_text = token.token_text.split(ParserHelper.newline_character)

        for split_index, next_text in enumerate(split_text):
            combined_text = (
                f"{split_whitespace[split_index + self.__paragraph_index]}{next_text}"
            )
            if (
                self.__first_line_after_hard_break
                or self.__first_line_after_other_token
                or split_index
            ):
                adjusted_column_number = (
                    self.__paragraph_column_number
                    if self.__first_line_after_other_token
                    else self.__paragraph_column_number
                    + len(split_whitespace[split_index + self.__paragraph_index])
                )

                if split_index == len(split_text) - 1:
                    self.__delayed_line = (
                        combined_text,
                        context,
                        token,
                        split_index,
                        -adjusted_column_number,
                    )
                else:
                    self.check_start_of_line(
                        combined_text,
                        context,
                        token,
                        split_index,
                        -adjusted_column_number,
                    )
            self.__first_line_after_other_token = False
            self.__first_line_after_hard_break = False
        self.__paragraph_index += token.token_text.count(ParserHelper.newline_character)

    # pylint: disable=too-many-arguments
    def check_start_of_line(
        self,
        combined_text: str,
        context: Any,
        token: MarkdownToken,
        line_number_delta: int,
        column_number_delta: int,
    ) -> None:
        """
        Check for a pattern at the start of the line.
        """

    # pylint: enable=too-many-arguments


class MyStartOfLineTokenParser(StartOfLineTokenParser):
    """
    Local implementation of the token parser.
    """

    def __init__(self, owner: "RuleMd018") -> None:
        super().__init__()
        self.__owner = owner

    # pylint: disable=too-many-arguments
    def check_start_of_line(
        self,
        combined_text: str,
        context: PluginScanContext,
        token: MarkdownToken,
        line_number_delta: int,
        column_number_delta: int,
    ) -> None:
        """
        Check for a pattern at the start of the line.

        Note that the logic here is a bit weird for the `\\S` at the end of the
        regular expression.  If everything else is good except for that `\\S`,
        which is equivalent to any whitespace character, and it was a space character,
        it would match the Atx code in the leaf processor and be an Atx Heading token
        already.  Hence, what is left is any other whitespace character which is not
        the space character, which should still trigger the rule.
        """
        if re.search(r"^[ ]{0,3}#{1,6}[^ ]", combined_text) and not re.search(
            r"#[ ]*$", combined_text
        ):
            self.__owner.report_next_token_error(
                context,
                token,
                line_number_delta=line_number_delta
                + context.calc_pragma_offset(token, line_number_delta),
                column_number_delta=column_number_delta,
            )

    # pylint: enable=too-many-arguments


class RuleMd018(RulePlugin):
    """
    Class to implement a plugin that looks for text in a paragraph where a line starts
    with what could be an atx heading, except there is no spaces between the hashes and
    the text of the heading.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__token_parser = MyStartOfLineTokenParser(self)

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-missing-space-atx",
            plugin_id="MD018",
            plugin_enabled_by_default=True,
            plugin_description="No space present after the hash character on a possible Atx Heading.",
            plugin_version="0.5.1",
            plugin_interface_version=1,
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md018.md",
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__token_parser.starting_new_file()

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        self.__token_parser.next_token(context, token)
